import os
import chevron


def prepare_instance_profile_template(managed_policy_arns: list):
    with open(os.path.join(os.path.dirname(__file__), 'data', 'template.yaml')) as f:
        content = f.read()

    parameters = {
        'HAS_MANAGED_POLICIES': len(managed_policy_arns),
        'MANAGED_POLICY_ARNS': [{'MANAGED_POLICY_ARN': arn} for arn in managed_policy_arns]
    }

    template = chevron.render(content, parameters)

    return template



import logging
from abc import ABC, abstractmethod
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.config.project_config import ProjectConfig
from spotty.deployment.abstract_cloud_instance.abstract_data_transfer import AbstractDataTransfer
from spotty.deployment.abstract_cloud_instance.abstract_instance_deployment import AbstractInstanceDeployment
from spotty.deployment.abstract_cloud_instance.abstract_bucket_manager import AbstractBucketManager
from spotty.deployment.abstract_cloud_instance.errors.bucket_not_found import BucketNotFoundError
from spotty.errors.instance_not_running import InstanceNotRunningError
from spotty.deployment.abstract_ssh_instance_manager import AbstractSshInstanceManager


class AbstractCloudInstanceManager(AbstractSshInstanceManager, ABC):

    def __init__(self, project_config: ProjectConfig, instance_config: dict):
        super().__init__(project_config, instance_config)

        self._bucket_manager = self._get_bucket_manager()
        self._data_transfer = self._get_data_transfer()
        self._instance_deployment = self._get_instance_deployment()

    @abstractmethod
    def _get_bucket_manager(self) -> AbstractBucketManager:
        """Returns an bucket manager."""
        raise NotImplementedError

    @abstractmethod
    def _get_data_transfer(self) -> AbstractDataTransfer:
        """Returns a data transfer object."""
        raise NotImplementedError

    @abstractmethod
    def _get_instance_deployment(self) -> AbstractInstanceDeployment:
        """Returns an instance deployment manager."""
        raise NotImplementedError

    @property
    def bucket_manager(self) -> AbstractBucketManager:
        """Returns a bucket manager."""
        return self._bucket_manager

    @property
    def data_transfer(self) -> AbstractDataTransfer:
        """Returns a data transfer object."""
        return self._data_transfer

    @property
    def instance_deployment(self) -> AbstractInstanceDeployment:
        """Returns an instance deployment manager."""
        return self._instance_deployment

    def is_running(self) -> bool:
        """Checks if the instance is running."""
        instance = self.instance_deployment.get_instance()
        return instance and instance.is_running

    def start(self, output: AbstractOutputWriter, dry_run=False):
        # make sure the Dockerfile exists
        self._check_dockerfile_exists()

        if not dry_run:
            # check if the instance is already running
            instance = self.instance_deployment.get_instance()
            if instance:
                if instance.is_running:
                    print('Instance is already running. Are you sure you want to restart it?')
                    res = input('Type "y" to confirm: ')
                    if res != 'y':
                        raise ValueError('The operation was cancelled.')

                    # terminating the instance to make EBS volumes available (the stack will be deleted later)
                    output.write('Terminating the instance... ', newline=False)
                    instance.terminate()
                    output.write('DONE')

                elif instance.is_stopped:
                    # TODO: restart the instance if it stopped
                    pass

        # create or get existing bucket for the project
        bucket_name = None
        try:
            bucket_name = self.bucket_manager.get_bucket().name
        except BucketNotFoundError:
            if not dry_run:
                bucket_name = self.bucket_manager.create_bucket().name
                output.write('Bucket "%s" was created.' % bucket_name)

        # deploy the instance
        self.instance_deployment.deploy(
            container_commands=self.container_commands,
            bucket_name=bucket_name,
            data_transfer=self.data_transfer,
            output=output,
            dry_run=dry_run,
        )

    def stop(self, only_shutdown: bool, output: AbstractOutputWriter):
        if only_shutdown:
            output.write('Shutting down the instance... ', newline=False)
            self.instance_deployment.get_instance().stop()
            output.write('DONE')
        else:
            # delete the stack and apply deletion policies
            self.instance_deployment.delete(output)

    def clean(self, output: AbstractOutputWriter):
        pass

    def sync(self, output: AbstractOutputWriter, dry_run=False):
        # get the project bucket name
        bucket_name = self.bucket_manager.get_bucket().name

        # sync the project with the S3 bucket
        output.write('Syncing the project with the bucket...')
        self.data_transfer.upload_local_to_bucket(bucket_name, dry_run=dry_run)

        if not dry_run:
            # sync the S3 bucket with the instance
            output.write('Syncing the bucket with the instance...')
            remote_cmd = self.data_transfer.get_download_bucket_to_instance_command(
                bucket_name=bucket_name,
                use_sudo=(not self.instance_config.container_config.run_as_host_user),
            )
            logging.debug('Remote sync command: ' + remote_cmd)

            # execute the command on the host OS
            exit_code = self.exec(remote_cmd)
            if exit_code != 0:
                raise ValueError('Failed to download files from the bucket to the instance')

    def download(self, download_filters: list, output: AbstractOutputWriter, dry_run=False):
        # get the project bucket name
        bucket_name = self.bucket_manager.get_bucket().name

        # sync files from the instance to a temporary S3 directory
        output.write('Uploading files from the instance to the bucket...')
        remote_cmd = self.data_transfer.get_upload_instance_to_bucket_command(
            bucket_name=bucket_name,
            download_filters=download_filters,
            use_sudo=(not self.instance_config.container_config.run_as_host_user),
            dry_run=dry_run,
        )
        logging.debug('Remote sync command: ' + remote_cmd)

        # execute the command on the host OS
        exit_code = self.exec(remote_cmd)
        if exit_code != 0:
            raise ValueError('Failed to upload files from the instance to the bucket')

        if not dry_run:
            # sync the project with the S3 bucket
            output.write('Downloading files from the bucket to local...')
            self.data_transfer.download_bucket_to_local(bucket_name=bucket_name, download_filters=download_filters)

    @property
    def ssh_host(self):
        """Returns an IP address that will be used for SSH connections."""
        if self._instance_config.local_ssh_port:
            return '127.0.0.1'

        # get a public IP address of the running instance
        instance = self.instance_deployment.get_instance()
        if not instance or not instance.is_running:
            raise InstanceNotRunningError(self.instance_config.name)

        public_ip_address = instance.public_ip_address
        if not public_ip_address:
            raise ValueError('The running instance doesn\'t have a public IP address.\n'
                             'Use the "localSshPort" parameter if you want to create a tunnel to the instance.')

        return public_ip_address

    @property
    def ssh_port(self) -> int:
        if self._instance_config.local_ssh_port:
            return self._instance_config.local_ssh_port

        return 22

    @property
    def use_tmux(self) -> bool:
        return True



from abc import ABC
from spotty.deployment.container.docker.docker_commands import DockerCommands
from spotty.deployment.container.abstract_container_script import AbstractContainerScript


class AbstractDockerScript(AbstractContainerScript, ABC):

    @property
    def commands(self) -> DockerCommands:
        return self._commands



import os
from schema import Schema, Optional, And, Regex, Or, Use
from spotty.config.validation import validate_config, get_instance_parameters_schema, has_prefix
from spotty.providers.gcp.config.image_uri import IMAGE_URI_REGEX


def validate_instance_parameters(params: dict):
    from spotty.providers.gcp.config.disk_volume import DiskVolume

    instance_parameters = {
        'zone': And(str, Regex(r'^[a-z0-9-]+$')),
        'machineType': str,
        Optional('gpu', default=None): {
            'type': str,
            Optional('count', default=1): int,
        },
        Optional('preemptibleInstance', default=False): bool,
        Optional('imageName', default=None): And(str, len, Regex(r'^[\w-]+$')),
        Optional('imageUri', default=None): And(str, len, Regex(IMAGE_URI_REGEX)),
        Optional('bootDiskSize', default=0): And(Or(int, str), Use(str),
                                                 Regex(r'^\d+$', error='Incorrect value for "bootDiskSize".'),
                                                 Use(int),
                                                 And(lambda x: x > 0,
                                                     error='"rootVolumeSize" should be greater than 0 or should '
                                                           'not be specified.'),
                                                 ),
        Optional('ports', default=[]): [And(int, lambda x: 0 < x < 65536)],
    }

    instance_checks = [
        And(lambda x: not (x['imageName'] and x['imageUri']),
            error='"imageName" and "imageUri" parameters cannot be used together.'),
    ]

    volume_checks = [
        And(lambda x: not has_prefix([(volume['parameters']['mountDir'] + '/') for volume in x
                                      if volume['parameters'].get('mountDir')]),
            error='Mount directories cannot be prefixes for each other.'),
    ]

    schema = get_instance_parameters_schema(instance_parameters, DiskVolume.TYPE_NAME, instance_checks, volume_checks)

    return validate_config(schema, params)


def validate_disk_volume_parameters(params: dict):
    from spotty.providers.gcp.config.disk_volume import DiskVolume

    schema = Schema({
        Optional('diskName', default=''): And(str, Regex(r'^[\w-]{1,255}$')),
        Optional('mountDir', default=''): And(
            str,
            And(os.path.isabs, error='Use absolute paths in the "mountDir" parameters'),
            Use(lambda x: x.rstrip('/'))
        ),
        Optional('size', default=0): And(int, lambda x: x > 0),
        Optional('deletionPolicy', default=DiskVolume.DP_RETAIN): And(
            str,
            lambda x: x in [DiskVolume.DP_CREATE_SNAPSHOT,
                            DiskVolume.DP_UPDATE_SNAPSHOT,
                            DiskVolume.DP_RETAIN,
                            DiskVolume.DP_DELETE], error='Incorrect value for "deletionPolicy".'
        ),
    })

    return validate_config(schema, params)



import logging
import subprocess
from spotty.deployment.abstract_cloud_instance.abstract_data_transfer import AbstractDataTransfer
from spotty.providers.aws.helpers.s3_sync import get_s3_sync_command, check_aws_installed


class DataTransfer(AbstractDataTransfer):

    def __init__(self, local_project_dir: str, host_project_dir: str, sync_filters: list, instance_name: str,
                 region: str):
        super().__init__(local_project_dir, host_project_dir, sync_filters, instance_name)

        self._region = region

    @property
    def scheme_name(self) -> str:
        return 's3'

    def upload_local_to_bucket(self, bucket_name: str, dry_run: bool = False):
        """Uploads files from local to the bucket."""
        # check AWS CLI is installed
        check_aws_installed()

        # sync the project with S3, deleted files will be deleted from S3
        local_cmd = get_s3_sync_command(self._local_project_dir, self._get_bucket_project_path(bucket_name),
                                        region=self._region, filters=self._sync_filters, delete=True, dry_run=dry_run)

        # execute the command locally
        logging.debug('Local sync command: ' + local_cmd)
        exit_code = subprocess.call(local_cmd, shell=True)
        if exit_code != 0:
            raise ValueError('Failed to upload the project files to the S3 bucket.')

    def download_bucket_to_local(self, bucket_name: str, download_filters: list):
        """Downloads files from the bucket to local."""
        # check AWS CLI is installed
        check_aws_installed()

        # download files from S3 bucket to local
        local_cmd = get_s3_sync_command(self._get_bucket_downloads_path(bucket_name), self._local_project_dir,
                                        region=self._region, filters=download_filters, exact_timestamp=True)

        # execute the command locally
        logging.debug('Local sync command: ' + local_cmd)
        exit_code = subprocess.call(local_cmd, shell=True)
        if exit_code != 0:
            raise ValueError('Failed to download files from the S3 bucket to local')

    def get_download_bucket_to_instance_command(self, bucket_name: str, use_sudo: bool = False) -> str:
        """A remote command to download files from the bucket to the instance."""
        remote_cmd = get_s3_sync_command(self._get_bucket_project_path(bucket_name), self._host_project_dir,
                                         region=self._region, filters=self._sync_filters, exact_timestamp=True)
        if use_sudo:
            remote_cmd = 'sudo ' + remote_cmd

        return remote_cmd

    def get_upload_instance_to_bucket_command(self, bucket_name: str, download_filters: list, use_sudo: bool = False,
                                              dry_run: bool = False) -> str:
        """A remote command to upload files from the instance to the bucket.

        It uses a temporary S3 directory that is unique for the instance. This
        directory keeps all downloaded from the instance files to sync only changed
        files with local.
        """

        # "sudo" should be called with the "-i" flag to use the root environment, so aws-cli will read
        # the config file from the root home directory
        remote_cmd = get_s3_sync_command(self._host_project_dir, self._get_bucket_downloads_path(bucket_name),
                                         region=self._region, filters=download_filters, delete=True, dry_run=dry_run)
        if use_sudo:
            remote_cmd = 'sudo ' + remote_cmd

        return remote_cmd



import os
import chevron
from spotty.deployment.container.docker.scripts.abstract_docker_script import AbstractDockerScript


class StopContainerScript(AbstractDockerScript):

    def render(self) -> str:
        # read template file
        template_path = os.path.join(os.path.dirname(__file__), 'data', 'stop_container.sh.tpl')
        with open(template_path) as f:
            template = f.read()

        # render the script
        content = chevron.render(template, data={
            'is_created_cmd': self.commands.is_created(),
            'remove_cmd': self.commands.remove(),
        })

        return content



from spotty.deployment.container.docker.scripts.start_container_script import StartContainerScript


class StartContainerScriptWithCfnSignals(StartContainerScript):

    @staticmethod
    def _get_signal_command(resource_name: str):
        return 'cfn-signal -e 0 --stack $_{AWS::StackName} --region $_{AWS::Region} --resource ' + resource_name

    def _partials(self) -> dict:
        return {
            'before_image_build': self._get_signal_command('BuildingDockerImageSignal'),
            'before_container_run': self._get_signal_command('StartingContainerSignal'),
            'before_startup_commands': self._get_signal_command('RunningContainerStartupCommandsSignal'),
        }

    def render(self, print_trace: bool = False) -> str:
        content = super().render(print_trace=print_trace)
        content = content.replace('${', '${!')
        content = content.replace('$_{', '${')

        return content



"""
INSTANCE FILE STRUCTURE
"""

# a base temporary directory on an instance
INSTANCE_SPOTTY_TMP_DIR = '/tmp/spotty'

# a base directory for container-related files and directories
CONTAINERS_TMP_DIR = INSTANCE_SPOTTY_TMP_DIR + '/containers'

# a base directory for instance-related files and directories
INSTANCE_DIR = INSTANCE_SPOTTY_TMP_DIR + '/instance'

# helper scripts
INSTANCE_SCRIPTS_DIR = INSTANCE_DIR + '/scripts'

# instance startup scripts
INSTANCE_STARTUP_SCRIPTS_DIR = INSTANCE_SCRIPTS_DIR + '/startup'

# a path to the script that attaches user to the container
CONTAINER_BASH_SCRIPT_PATH = INSTANCE_SCRIPTS_DIR + '/container_bash.sh'



import os
from abc import ABC
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.deployment.utils.commands import get_script_command
from spotty.deployment.container.docker.docker_commands import DockerCommands
from spotty.deployment.container.docker.scripts.start_container_script import StartContainerScript
from spotty.deployment.container.docker.scripts.stop_container_script import StopContainerScript
from spotty.deployment.abstract_instance_manager import AbstractInstanceManager
from spotty.errors.nothing_to_do import NothingToDoError
from spotty.utils import render_table


class AbstractDockerInstanceManager(AbstractInstanceManager, ABC):

    @property
    def container_commands(self) -> DockerCommands:
        """A collection of commands to manage a container from the host OS."""
        return DockerCommands(self.instance_config)

    def is_container_running(self) -> bool:
        """Checks if the container is running."""
        is_running_cmd = self.container_commands.is_created(is_running=True)
        exit_code = self.exec(is_running_cmd)

        return exit_code == 0

    def start_container(self, output: AbstractOutputWriter, dry_run=False):
        """Starts or restarts container on the host OS."""
        # make sure the Dockerfile exists
        self._check_dockerfile_exists()

        # sync the project with the instance
        try:
            self.sync(output, dry_run=dry_run)
        except NothingToDoError:
            pass

        # generate a script that starts container
        start_container_script = StartContainerScript(self.container_commands).render()
        start_container_command = get_script_command('start-container', start_container_script)

        # start the container
        exit_code = self.exec(start_container_command)
        if exit_code != 0:
            raise ValueError('Failed to start the container')

    def start(self, output: AbstractOutputWriter, dry_run=False):
        # start or restart container
        self.start_container(output, dry_run=dry_run)

    def stop(self, only_shutdown: bool, output: AbstractOutputWriter):
        # stop container
        stop_container_script = StopContainerScript(self.container_commands).render()
        stop_container_command = get_script_command('stop-container', stop_container_script)

        exit_code = self.exec(stop_container_command)
        if exit_code != 0:
            raise ValueError('Failed to stop the container')

    def get_status_text(self):
        if self.is_container_running():
            msg = 'Container is running.'
        else:
            msg = 'Container is not running.'

        return render_table([(msg,)])

    def _check_dockerfile_exists(self):
        """Raises an error if a Dockerfile specified in the configuration file but doesn't exist."""
        if self.instance_config.container_config.file:
            dockerfile_path = os.path.join(self.project_config.project_dir, self.instance_config.container_config.file)
            if not os.path.isfile(dockerfile_path):
                raise FileNotFoundError('A Dockerfile specified in the container configuration doesn\'t exist:\n  ' +
                                        dockerfile_path)



from shutil import which
from spotty.deployment.utils.cli import shlex_join


def check_aws_installed():
    """Checks that AWS CLI is installed."""
    if which('aws') is None:
        raise ValueError('AWS CLI is not installed.')


def get_s3_sync_command(from_path: str, to_path: str, profile: str = None, region: str = None, filters: list = None,
                        exact_timestamp: bool = False, delete: bool = False, quiet: bool = False,
                        dry_run: bool = False):
    """Builds an "aws s3 sync" command."""
    args = ['aws']

    if profile:
        args += ['--profile', profile]

    if region:
        args += ['--region', region]

    args += ['s3', 'sync', from_path, to_path]

    if filters:
        for sync_filter in filters:
            if ('exclude' in sync_filter and 'include' in sync_filter) \
                    or ('exclude' not in sync_filter and 'include' not in sync_filter):
                raise ValueError('S3 sync filter has wrong format.')

            if 'exclude' in sync_filter:
                for path in sync_filter['exclude']:
                    args += ['--exclude', path]

            if 'include' in sync_filter:
                for path in sync_filter['include']:
                    args += ['--include', path]

    if exact_timestamp:
        args.append('--exact-timestamp')

    if delete:
        args.append('--delete')

    if quiet:
        args.append('--quiet')

    if dry_run:
        args.append('--dryrun')

    command = shlex_join(args)

    return command



from argparse import Namespace
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.deployment.abstract_instance_manager import AbstractInstanceManager


class StopCommand(AbstractConfigCommand):

    name = 'stop'
    description = 'Terminate running instance and apply deletion policies for the volumes'

    # TODO: the "spotty start" command should restart the instance and the container if the instance was shutdown
    # def configure(self, parser: ArgumentParser):
    #     super().configure(parser)
    #     parser.add_argument('-s', '--shutdown', action='store_true',
    #                         help='Shutdown the instance without terminating it. Deletion policies for the volumes '
    #                              'won\'t be applied.')

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):

        instance_manager.stop(only_shutdown=False, output=output)



from collections import OrderedDict
from time import sleep
import googleapiclient.discovery


class CEClient(object):
    """Compute Engine client."""

    def __init__(self, project_id: str, zone: str):
        self._project_id = project_id
        self._zone = zone
        self._client = googleapiclient.discovery.build('compute', 'v1', cache_discovery=False)

    @property
    def zone(self):
        return self._zone

    def list_images(self, image_name: str = None, project_id: str = None):
        """Returns a list of images that satisfy the name.
            This method is used instead of the "get" because it doesn't raise an exception if an image doesn't exist.
        """
        if not project_id:
            project_id = self._project_id

        filter_str = ('name=%s' % image_name) if image_name else None
        res = self._client.images().list(project=project_id, filter=filter_str).execute()

        if not res.get('items'):
            return []

        return res['items']

    def get_image_from_family(self, family_name: str, project_id: str = None):
        if not project_id:
            project_id = self._project_id

        res = self._client.images().getFromFamily(project=project_id, family=family_name).execute()

        return res

    def list_instances(self, machine_name=None):
        filter_str = ('name=%s' % machine_name) if machine_name else None
        res = self._client.instances().list(project=self._project_id, zone=self._zone, filter=filter_str).execute()

        if not res.get('items'):
            return []

        return res['items']

    def list_disks(self, disk_name=None):
        filter_str = ('name=%s' % disk_name) if disk_name else None
        res = self._client.disks().list(project=self._project_id, zone=self._zone, filter=filter_str).execute()

        if not res.get('items'):
            return []

        return res['items']

    def list_snapshots(self, snapshot_name=None):
        filter_str = ('name=%s' % snapshot_name) if snapshot_name else None
        res = self._client.snapshots().list(project=self._project_id, filter=filter_str).execute()

        if not res.get('items'):
            return []

        return res['items']

    def get_accelerator_types(self) -> OrderedDict:
        res = self._client.acceleratorTypes().list(project=self._project_id, zone=self._zone).execute()
        accelerator_types = OrderedDict([(item['name'], item['maximumCardsPerInstance'])
                                         for item in res.get('items', [])])

        return accelerator_types

    def create_disk(self, name: str, size: int = None, snapshot_link: str = None) -> str:
        params = {
            'name': name,
            'type': 'zones/%s/diskTypes/pd-standard' % self._zone,
            'physicalBlockSizeBytes': 4096,
        }

        if size:
            params['sizeGb'] = size

        if snapshot_link:
            params['sourceSnapshot'] = snapshot_link

        res = self._client.disks().insert(project=self._project_id, zone=self._zone, body=params).execute()

        return res['targetLink']

    def get_machine_types(self, machine_type: str = None):
        """Returns a list of images that satisfy the name.
            This method is used instead of the "get" because it doesn't raise an exception if an image doesn't exist.
        """
        filter_str = ('name=%s' % machine_type) if machine_type else None
        res = self._client.machineTypes().list(project=self._project_id, zone=self._zone, filter=filter_str).execute()

        if not res.get('items'):
            return []

        return res['items']

    def stop_instance(self, machine_name: str, wait: bool = True) -> str:
        """Stops the instance."""
        operation = self._client.instances().stop(project=self._project_id, zone=self._zone,
                                            instance=machine_name).execute()
        if wait:
            operation = self._wait_operation(operation)

        return operation['targetLink']

    def delete_instance(self, machine_name: str, wait: bool = True) -> str:
        """Deletes the instance."""
        operation = self._client.instances().delete(project=self._project_id, zone=self._zone,
                                                    instance=machine_name).execute()
        if wait:
            operation = self._wait_operation(operation)

        return operation['targetLink']

    def _wait_operation(self, operation: dict):
        """Waits util the operation is finished/"""
        while operation['status'] != 'DONE':
            sleep(5)
            operation = self._client.zoneOperations().wait(project=self._project_id, zone=self._zone,
                                                           operation=operation['name']).execute()

        return operation



import os
import chevron
from spotty.deployment.utils.commands import get_bash_command
from spotty.deployment.container.docker.scripts.abstract_docker_script import AbstractDockerScript


class ContainerBashScript(AbstractDockerScript):

    def render(self) -> str:
        # read template file
        template_path = os.path.join(os.path.dirname(__file__), 'data', 'container_bash.sh.tpl')
        with open(template_path) as f:
            template = f.read()

        # render the script
        content = chevron.render(template, data={
            'docker_exec_bash': self.commands.exec(get_bash_command(), interactive=True, tty=True,
                                                   container_name='$SPOTTY_CONTAINER_NAME',
                                                   working_dir='$SPOTTY_CONTAINER_WORKING_DIR'),
        })

        return content



from typing import List
import os
import chevron
import yaml
from cfn_tools import CfnYamlLoader, CfnYamlDumper
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.config.tmp_dir_volume import TmpDirVolume
from spotty.config.validation import is_subdir
from spotty.config.abstract_instance_volume import AbstractInstanceVolume
from spotty.deployment.container.docker.docker_commands import DockerCommands
from spotty.deployment.container.docker.scripts.container_bash_script import ContainerBashScript
from spotty.deployment.abstract_cloud_instance.file_structure import INSTANCE_SPOTTY_TMP_DIR, \
    CONTAINER_BASH_SCRIPT_PATH, \
    INSTANCE_STARTUP_SCRIPTS_DIR, CONTAINERS_TMP_DIR
from spotty.providers.aws.cfn_templates.instance.start_container_script import StartContainerScriptWithCfnSignals
from spotty.providers.aws.helpers.ami import get_ami
from spotty.providers.aws.helpers.vpc import get_vpc_id
from spotty.providers.aws.resources.snapshot import Snapshot
from spotty.providers.aws.resources.volume import Volume
from spotty.providers.aws.config.instance_config import InstanceConfig
from spotty.providers.aws.config.ebs_volume import EbsVolume
from spotty.providers.aws.helpers.logs import get_logs_s3_path


def prepare_instance_template(ec2, instance_config: InstanceConfig, docker_commands: DockerCommands,
                              availability_zone: str, sync_project_cmd: str, output: AbstractOutputWriter):
    """Prepares CloudFormation template to run a Spot Instance."""

    # read and update CF template
    with open(os.path.join(os.path.dirname(__file__), 'data', 'template.yaml')) as f:
        template = yaml.load(f, Loader=CfnYamlLoader)

    # get volume resources and updated availability zone
    volume_resources = _get_volume_resources(ec2, instance_config.volumes, output)

    # add volume resources to the template
    template['Resources'].update(volume_resources)

    # set availability zone
    if availability_zone:
        template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['Placement'] = {
            'AvailabilityZone': availability_zone,
        }
        output.write('- availability zone: %s' % availability_zone)
    else:
        output.write('- availability zone: auto')

    # set subnet
    if instance_config.subnet_id:
        template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['NetworkInterfaces'] = [
            {
                'SubnetId': instance_config.subnet_id,
                'DeviceIndex': 0,
                'Groups': template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData'][
                    'SecurityGroupIds'],
            }]
        del template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['SecurityGroupIds']

    # add ports to the security group
    for port in instance_config.ports:
        if port != 22:
            template['Resources']['InstanceSecurityGroup']['Properties']['SecurityGroupIngress'] += [{
                'CidrIp': '0.0.0.0/0',
                'IpProtocol': 'tcp',
                'FromPort': port,
                'ToPort': port,
            }, {
                'CidrIpv6': '::/0',
                'IpProtocol': 'tcp',
                'FromPort': port,
                'ToPort': port,
            }]

    if instance_config.is_spot_instance:
        # set maximum price
        if instance_config.max_price:
            template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData'] \
                ['InstanceMarketOptions']['SpotOptions']['MaxPrice'] = instance_config.max_price

        output.write('- maximum Spot Instance price: %s'
                     % (('%.04f' % instance_config.max_price) if instance_config.max_price else 'on-demand'))
    else:
        # run on-demand instance
        del template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData'][
            'InstanceMarketOptions']
        output.write('- on-demand instance')

    # set the user data script
    template['Resources']['InstanceLaunchTemplate']['Properties']['LaunchTemplateData']['UserData'] = {
        'Fn::Base64': {
            'Fn::Sub': _read_template_file(os.path.join('startup_scripts', 'user_data.sh')),
        },
    }

    # run sync command as a non-root user
    if instance_config.container_config.run_as_host_user:
        sync_project_cmd = 'sudo -u %s %s' % (instance_config.user, sync_project_cmd)

    # get mount directories
    mount_dirs = [volume.mount_dir for volume in instance_config.volumes if isinstance(volume, EbsVolume)]

    # set CloudFormation configs
    cfn_init_configs = [
        {
            'name': 'prepare_instance',
            'files': {
                INSTANCE_STARTUP_SCRIPTS_DIR + '/01_prepare_instance.sh': {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000755',
                    'content': {
                        'Fn::Sub': _read_template_file(os.path.join('startup_scripts', '01_prepare_instance.sh'), {
                            'CONTAINER_BASH_SCRIPT_PATH': CONTAINER_BASH_SCRIPT_PATH,
                            'SPOTTY_TMP_DIR': INSTANCE_SPOTTY_TMP_DIR,
                            'CONTAINERS_TMP_DIR': CONTAINERS_TMP_DIR,
                        }),
                    },
                },
                CONTAINER_BASH_SCRIPT_PATH: {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000755',
                    'content': ContainerBashScript(docker_commands).render(),
                },
                '/home/ubuntu/.tmux.conf': {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000644',
                    'content': {
                        'Fn::Sub': _read_template_file(os.path.join('files', 'tmux.conf')),
                    },
                },
            },
            'command': INSTANCE_STARTUP_SCRIPTS_DIR + '/01_prepare_instance.sh',
        },
        {
            'name': 'mount_volumes',
            'files': {
                INSTANCE_STARTUP_SCRIPTS_DIR + '/02_mount_volumes.sh': {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000755',
                    'content': {
                        'Fn::Sub': _read_template_file(os.path.join('startup_scripts', '02_mount_volumes.sh'), {
                            'MOUNT_DIRS': ('"%s"' % '" "'.join(mount_dirs)) if mount_dirs else '',
                            'TMP_VOLUME_DIRS': [{'PATH': volume.host_path} for volume in instance_config.volumes
                                                if isinstance(volume, TmpDirVolume)],
                        }),
                    },
                },
            },
            'command': INSTANCE_STARTUP_SCRIPTS_DIR + '/02_mount_volumes.sh',
        },
        {
            'name': 'set_docker_root',
            'files': {
                INSTANCE_STARTUP_SCRIPTS_DIR + '/03_set_docker_root.sh': {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000755',
                    'content': {
                        'Fn::Sub': _read_template_file(os.path.join('startup_scripts', '03_set_docker_root.sh')),
                    },
                },
            },
            'command': INSTANCE_STARTUP_SCRIPTS_DIR + '/03_set_docker_root.sh',
        },
        {
            'name': 'sync_project',
            'files': {
                INSTANCE_STARTUP_SCRIPTS_DIR + '/04_sync_project.sh': {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000755',
                    'content': {
                        'Fn::Sub': _read_template_file(os.path.join('startup_scripts', '04_sync_project.sh'), {
                            'SYNC_PROJECT_CMD': sync_project_cmd,
                        }),
                    },
                },
            },
            'command': INSTANCE_STARTUP_SCRIPTS_DIR + '/04_sync_project.sh',
        },
        {
            'name': 'run_instance_startup_commands',
            'files': {
                INSTANCE_STARTUP_SCRIPTS_DIR + '/05_run_instance_startup_commands.sh': {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000755',
                    'content': {
                        'Fn::Sub': _read_template_file(
                            os.path.join('startup_scripts', '05_run_instance_startup_commands.sh'), {
                                'INSTANCE_STARTUP_SCRIPTS_DIR': INSTANCE_STARTUP_SCRIPTS_DIR,
                            }),
                    },
                },
                INSTANCE_STARTUP_SCRIPTS_DIR + '/instance_startup_commands.sh': {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000644',
                    'content': instance_config.commands or '#',
                },
            },
            'command': INSTANCE_STARTUP_SCRIPTS_DIR + '/05_run_instance_startup_commands.sh',
        },
        {
            'name': 'start_container',
            'files': {
                INSTANCE_STARTUP_SCRIPTS_DIR + '/06_start_container.sh': {
                    'owner': 'ubuntu',
                    'group': 'ubuntu',
                    'mode': '000755',
                    'content': {
                        'Fn::Sub': StartContainerScriptWithCfnSignals(docker_commands).render(print_trace=True),
                    },
                },
            },
            'command': INSTANCE_STARTUP_SCRIPTS_DIR + '/06_start_container.sh',
        },
    ]

    template['Resources']['InstanceLaunchTemplate']['Metadata']['AWS::CloudFormation::Init']['configSets'] = {
        'init': [config['name'] for config in cfn_init_configs],
    }

    for config in cfn_init_configs:
        template['Resources']['InstanceLaunchTemplate']['Metadata']['AWS::CloudFormation::Init'][config['name']] = {
            'files': config.get('files', {}),
            'commands': {
                config['name']: {
                    'command': config['command'],
                }
            },
        }

    return yaml.dump(template, Dumper=CfnYamlDumper)


def _read_template_file(filename: str, params: dict = None):
    with open(os.path.join(os.path.dirname(__file__), 'data', filename)) as f:
        content = f.read()

    if params:
        content = chevron.render(content, params)

    return content


def _get_volume_attachment_resource(volume_id, device_name):
    attachment_resource = {
        'Type': 'AWS::EC2::VolumeAttachment',
        'Properties': {
            'Device': device_name,
            'InstanceId': {'Ref': 'Instance'},
            'VolumeId': volume_id if isinstance(volume_id, str) else dict(volume_id),  # avoid YAML aliases
        },
        'Metadata': {
            'Device': device_name,
            'VolumeId': volume_id if isinstance(volume_id, str) else dict(volume_id),  # avoid YAML aliases
        },
    }

    return attachment_resource


def _get_volume_resource(ec2, volume: EbsVolume, output: AbstractOutputWriter):
    # new volume will be created
    volume_resource = {
        'Type': 'AWS::EC2::Volume',
        'DeletionPolicy': 'Retain',
        'Properties': {
            'AvailabilityZone': {'Fn::GetAtt': ['Instance', 'AvailabilityZone']},
            'Tags': [{
                'Key': 'Name',
                'Value': volume.ec2_volume_name,
            }],
            'VolumeType': volume.type,
        },
    }

    # check if the snapshot exists and restore the volume from it
    snapshot = Snapshot.get_by_name(ec2, volume.ec2_volume_name)
    if snapshot:
        # volume will be restored from the snapshot
        # check size of the volume
        if volume.size and (volume.size < snapshot.size):
            raise ValueError('Specified size for the "%s" volume (%dGB) is less than size of the '
                             'snapshot (%dGB).'
                             % (volume.name, volume.size, snapshot.size))

        # set snapshot ID
        volume_resource['Properties']['SnapshotId'] = snapshot.snapshot_id

        output.write('- volume "%s" will be restored from the snapshot' % volume.ec2_volume_name)

    else:
        # empty volume will be created, check that the size is specified
        if not volume.size:
            raise ValueError('Size for the new volume is required.')

        output.write('- volume "%s" will be created' % volume.ec2_volume_name)

    # set size of the volume
    if volume.size:
        volume_resource['Properties']['Size'] = volume.size

    # set a name for the new volume
    volume_resource['Properties']['Tags'] = [{'Key': 'Name', 'Value': volume.ec2_volume_name}]

    return volume_resource


def _get_volume_resources(ec2, volumes: List[AbstractInstanceVolume], output: AbstractOutputWriter):
    resources = {}

    # ending letters for the devices (see: https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/device_naming.html)
    device_letters = 'fghijklmnop'

    # create and attach volumes
    for i, volume in enumerate(volumes):
        if isinstance(volume, EbsVolume):
            device_letter = device_letters[i]

            ec2_volume = Volume.get_by_name(ec2, volume.ec2_volume_name)
            if ec2_volume:
                # check if the volume is available
                if not ec2_volume.is_available():
                    raise ValueError('EBS volume "%s" is not available (state: %s).'
                                     % (volume.ec2_volume_name, ec2_volume.state))

                # check size of the volume
                if volume.size and (volume.size != ec2_volume.size):
                    raise ValueError('Specified size for the "%s" volume (%dGB) doesn\'t match the size of the '
                                     'existing volume (%dGB).' % (volume.name, volume.size, ec2_volume.size))

                output.write('- volume "%s" (%s) will be attached' % (ec2_volume.name, ec2_volume.volume_id))

                volume_id = ec2_volume.volume_id
            else:
                # create Volume resource
                vol_resource_name = 'Volume' + device_letter.upper()
                vol_resource = _get_volume_resource(ec2, volume, output)
                resources[vol_resource_name] = vol_resource

                volume_id = {'Ref': vol_resource_name}

            # create VolumeAttachment resource
            vol_attachment_resource_name = 'VolumeAttachment' + device_letter.upper()
            device_name = '/dev/sd' + device_letter
            vol_attachment_resource = _get_volume_attachment_resource(volume_id, device_name)
            resources[vol_attachment_resource_name] = vol_attachment_resource

    return resources


def get_template_parameters(ec2, instance_config: InstanceConfig, instance_profile_arn: str, bucket_name: str,
                            key_pair_name: str, output: AbstractOutputWriter):
    # get AMI
    ami = get_ami(ec2, instance_config.ami_id, instance_config.ami_name)
    output.write('- AMI: "%s" (%s)' % (ami.name, ami.image_id))

    # check root volume size
    root_volume_size = instance_config.root_volume_size
    if root_volume_size and root_volume_size < ami.size:
        raise ValueError('Root volume size cannot be less than the size of AMI (%dGB).' % ami.size)
    elif not root_volume_size:
        # if a root volume size is not specified, make it 5GB larger than the AMI size
        root_volume_size = ami.size + 5

    # print info about the Docker data root
    ebs_volumes = [volume for volume in instance_config.volumes if isinstance(volume, EbsVolume)]
    if instance_config.docker_data_root:
        docker_data_volume_name = [volume.name for volume in ebs_volumes
                                   if is_subdir(instance_config.docker_data_root, volume.mount_dir)][0]
        output.write('- Docker data will be stored on the "%s" volume' % docker_data_volume_name)

    # create stack
    parameters = {
        'VpcId': get_vpc_id(ec2, instance_config.subnet_id),
        'InstanceProfileArn': instance_profile_arn,
        'InstanceType': instance_config.instance_type,
        'KeyName': key_pair_name,
        'ImageId': ami.image_id,
        'RootVolumeSize': str(root_volume_size),
        'DockerDataRootDirectory': instance_config.docker_data_root,
        'InstanceNameTag': instance_config.ec2_instance_name,
        'HostProjectDirectory': instance_config.host_project_dir,
        'LogsS3Path': get_logs_s3_path(bucket_name, instance_config.name),
    }

    return parameters



from spotty.config.abstract_instance_volume import AbstractInstanceVolume
from spotty.providers.aws.config.validation import validate_ebs_volume_parameters


class EbsVolume(AbstractInstanceVolume):

    TYPE_NAME = 'EBS'

    DP_CREATE_SNAPSHOT = 'CreateSnapshot'
    DP_UPDATE_SNAPSHOT = 'UpdateSnapshot'
    DP_RETAIN = 'Retain'
    DP_DELETE = 'Delete'

    def __init__(self, volume_config: dict, project_name: str, instance_name: str):
        super().__init__(volume_config)

        self._project_name = project_name
        self._instance_name = instance_name

    def _validate_volume_parameters(self, params: dict) -> dict:
        return validate_ebs_volume_parameters(params)

    @property
    def title(self):
        return 'EBS volume'

    @property
    def size(self) -> int:
        return self._params['size']

    @property
    def type(self) -> str:
        return self._params['type']

    @property
    def deletion_policy(self) -> str:
        return self._params['deletionPolicy']

    @property
    def deletion_policy_title(self) -> str:
        return {
            EbsVolume.DP_CREATE_SNAPSHOT: 'Create Snapshot',
            EbsVolume.DP_UPDATE_SNAPSHOT: 'Update Snapshot',
            EbsVolume.DP_RETAIN: 'Retain Volume',
            EbsVolume.DP_DELETE: 'Delete Volume',
        }[self.deletion_policy]

    @property
    def ec2_volume_name(self) -> str:
        """Returns EBS volume name."""
        volume_name = self._params['volumeName']
        if not volume_name:
            volume_name = '%s-%s-%s' % (self._project_name.lower(), self._instance_name.lower(), self.name.lower())

        return volume_name

    @property
    def mount_dir(self) -> str:
        """A directory where the volume will be mounted on the host OS."""
        if self._params['mountDir']:
            mount_dir = self._params['mountDir']
        else:
            mount_dir = '/mnt/%s' % self.ec2_volume_name

        return mount_dir

    @property
    def host_path(self) -> str:
        """A path on the host OS that will be mounted to the container."""
        return self.mount_dir



class ImageNotFoundError(Exception):
    def __init__(self, image_name):
        super().__init__('The image "%s" was not found.\n'
                         'Use the "spotty gcp create-image" command to create an image with NVIDIA Docker.'
                         % image_name)



import os
from typing import List
from spotty.config.abstract_instance_config import AbstractInstanceConfig, VolumeMount
from spotty.config.abstract_instance_volume import AbstractInstanceVolume
from spotty.config.container_config import PROJECT_VOLUME_MOUNT_NAME
from spotty.config.project_config import ProjectConfig
from spotty.config.host_path_volume import HostPathVolume
from spotty.providers.local.config.validation import validate_instance_parameters


class InstanceConfig(AbstractInstanceConfig):

    def __init__(self, instance_config: dict, project_config: ProjectConfig):
        super().__init__(instance_config, project_config)

    def _validate_instance_params(self, params: dict):
        # validate the config and fill missing parameters with the default values
        return validate_instance_parameters(params)

    def _get_instance_volumes(self) -> List[AbstractInstanceVolume]:
        volumes = []
        for volume_config in self._params['volumes']:
            volume_type = volume_config['type']
            if volume_type == HostPathVolume.TYPE_NAME:
                volumes.append(HostPathVolume(volume_config, self.project_config.project_dir))
            else:
                raise ValueError('Volume type "%s" is not supported.' % volume_type)

        return volumes

    def _get_volume_mounts(self, volumes: List[AbstractInstanceVolume]) -> List[VolumeMount]:
        volume_mounts = super()._get_volume_mounts(volumes)

        # ignore a volume that matches the container project directory
        volume_mounts = [volume_mount for volume_mount in volume_mounts
                         if os.path.relpath(self.container_config.project_dir, volume_mount.mount_path) != '.']

        # mount the local project directory to the container
        volume_mounts.append(VolumeMount(
            name=PROJECT_VOLUME_MOUNT_NAME,
            host_path=self.project_config.project_dir,
            mount_path=self.container_config.project_dir,
            mode='rw',
            hidden=True,
        ))

        return volume_mounts

    @property
    def user(self) -> str:
        return ''



from typing import List
from spotty.config.abstract_instance_config import VolumeMount
from spotty.config.abstract_instance_volume import AbstractInstanceVolume
from spotty.config.container_config import PROJECT_VOLUME_MOUNT_NAME
from spotty.utils import render_table


def render_volumes_info_table(volume_mounts: List[VolumeMount], volumes: List[AbstractInstanceVolume]):
    table = [('Name', 'Mount Path', 'Type', 'Deletion Policy')]

    # add volume mounts to the info table
    volumes_dict = {volume.name: volume for volume in volumes}
    for volume_mount in volume_mounts:
        if not volume_mount.hidden:
            # the volume will be mounted to the container
            volume = volumes_dict[volume_mount.name]
            vol_mount_name = '-' if volume_mount.name == PROJECT_VOLUME_MOUNT_NAME else volume_mount.name
            deletion_policy = volume.deletion_policy_title if volume.deletion_policy_title else '-'
            table.append((vol_mount_name, volume_mount.mount_path, volume.title, deletion_policy))

    # add volumes that were not mounted to the container to the info table
    volume_mounts_dict = {volume_mount.name for volume_mount in volume_mounts}
    for volume in volumes:
        if volume.name not in volume_mounts_dict:
            deletion_policy = volume.deletion_policy_title if volume.deletion_policy_title else '-'
            table.append((volume.name, '-', volume.title, deletion_policy))

    return render_table(table, separate_title=True)



from spotty.providers.gcp.helpers.dm_client import DMClient


class DMResource(object):

    def __init__(self, dm: DMClient, data: dict):
        """
        Args:
            dm (DMClient): Deployment Manager client
            data (dict): Stack info.
                Example #1:
                {'id': '1760655646875625396',
                 'insertTime': '2019-08-25T16:27:23.544-07:00',
                 'name': 'x11-test-i2-docker-waiter',
                 'type': 'runtimeconfig.v1beta1.waiter',
                 'update': {'finalProperties': 'failure:\n'
                                               '  cardinality:\n'
                                               '    number: 1\n'
                                               '    path: /failure\n'
                                               'parent: '
                                               'projects/spotty-221422/configs/x11-test-i2-docker-status\n'
                                               'success:\n'
                                               '  cardinality:\n'
                                               '    number: 1\n'
                                               '    path: /success\n'
                                               'timeout: 1800s\n'
                                               'waiter: x11-test-i2-docker-waiter\n',
                            'intent': 'CREATE_OR_ACQUIRE',
                            'manifest': 'https://www.googleapis.com/deploymentmanager/v2/projects/spotty-221422/global/deployments/spotty-instance-x11-test-i2/manifests/manifest-1566775635906',
                            'properties': 'failure:\n'
                                          '  cardinality:\n'
                                          '    number: 1\n'
                                          '    path: /failure\n'
                                          'parent: $(ref.x11-test-i2-docker-status.name)\n'
                                          'success:\n'
                                          '  cardinality:\n'
                                          '    number: 1\n'
                                          '    path: /success\n'
                                          'timeout: 1800s\n'
                                          'waiter: x11-test-i2-docker-waiter\n',
                            'state': 'IN_PROGRESS'},
                 'updateTime': '2019-08-25T16:27:23.544-07:00'}

                Example #2:
                {'finalProperties': 'config: x11-test-i2-docker-status\n'
                                    'description: Docker status\n',
                 'id': '314866945194106123',
                 'insertTime': '2019-08-25T17:12:20.140-07:00',
                 'manifest': 'https://www.googleapis.com/deploymentmanager/v2/projects/spotty-221422/global/deployments/spotty-instance-x11-test-i2/manifests/manifest-1566778333272',
                 'name': 'x11-test-i2-docker-status',
                 'properties': 'config: x11-test-i2-docker-status\n'
                               'description: Docker status\n',
                 'type': 'runtimeconfig.v1beta1.config',
                 'updateTime': '2019-08-25T17:12:30.254-07:00',
                 'url': 'https://runtimeconfig.googleapis.com/v1beta1/projects/spotty-221422/configs/x11-test-i2-docker-status'}
        """
        self._dm = dm
        self._data = data

    @staticmethod
    def get_by_name(dm: DMClient, deployment_name: str, resource_name: str):
        """Returns an instance by its stack name."""
        res = dm.get_resource(deployment_name, resource_name)
        if not res:
            return None

        return DMResource(dm, res)

    @property
    def is_created(self) -> bool:
        return 'finalProperties' in self._data

    @property
    def error_message(self) -> str:
        if 'error' not in self._data.get('update', {}):
            return None

        return self._data['update']['error']['errors'][0]['message']

    @property
    def state(self) -> str:
        return self._data['update']['state'] if 'state' in self._data.get('update', {}) else None

    @property
    def is_in_progress(self) -> bool:
        return self.state == 'IN_PROGRESS'

    @property
    def is_failed(self) -> bool:
        # an error occurred or the resource is in an unexpected status
        return self.error_message or (self.state is not None and
                                      self.state not in ['PENDING', 'IN_PROGRESS', 'COMPLETED', 'IN_PREVIEW'])



import logging
import os
from abc import abstractmethod
from spotty.deployment.utils.commands import get_ssh_command
from spotty.deployment.abstract_docker_instance_manager import AbstractDockerInstanceManager


class AbstractSshInstanceManager(AbstractDockerInstanceManager):

    def exec(self, command: str, tty: bool = True) -> int:
        """Executes a command on the host OS."""
        if not os.path.isfile(self.ssh_key_path):
            raise ValueError('SSH key doesn\'t exist: ' + self.ssh_key_path)

        ssh_command = get_ssh_command(self.ssh_host, self.ssh_port, self.ssh_user, self.ssh_key_path,
                                      command, env_vars=self.ssh_env_vars, tty=tty)
        logging.debug('SSH command: ' + ssh_command)

        return super().exec(ssh_command)

    @property
    @abstractmethod
    def ssh_host(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def ssh_port(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def ssh_key_path(self) -> str:
        raise NotImplementedError

    @property
    def ssh_user(self) -> str:
        return self.instance_config.user

    @property
    def ssh_env_vars(self) -> dict:
        """Environmental variables that will be set when ssh to the instance."""
        return {
            'SPOTTY_CONTAINER_NAME': self.instance_config.full_container_name,
            'SPOTTY_CONTAINER_WORKING_DIR': self.instance_config.container_config.working_dir,
        }

    @property
    def use_tmux(self) -> bool:
        """Use tmux when running a custom script or connecting to the instance."""
        return True



from spotty.config.host_path_volume import HostPathVolume


class TmpDirVolume(HostPathVolume):

    @property
    def title(self):
        return 'temporary directory'

    @property
    def deletion_policy_title(self) -> str:
        return ''



import googleapiclient.discovery


class RtcClient(object):

    def __init__(self, project_id: str, zone: str):
        self._project_id = project_id
        self._zone = zone
        self._rtc = googleapiclient.discovery.build('runtimeconfig', 'v1beta1', cache_discovery=False)

    def get_value(self, config_name, template):
        config_name = 'projects/%s/configs/%s' % (self._project_id, config_name)
        fields = ['/failure']
        res = self._rtc.projects().get(name=config_name, fields=fields).execute()

        return res

    def set_value(self, config_name: str, variable_name: str, value: str):
        config_name = 'projects/%s/configs/%s' % (self._project_id, config_name)
        res = self._rtc.projects().configs().variables().create(parent=config_name, body={
            'name': '%s/variables/%s' % (config_name, variable_name),
            'text': str(value),
        }).execute()

        return res



import argparse
from typing import List, Type
import pkg_resources
from spotty.commands.abstract_command import AbstractCommand
from spotty.commands.aws import AwsCommand
from spotty.commands.download import DownloadCommand
from spotty.commands.exec import ExecCommand
from spotty.commands.run import RunCommand
from spotty.commands.sh import ShCommand
from spotty.commands.start import StartCommand
from spotty.commands.status import StatusCommand
from spotty.commands.stop import StopCommand
from spotty.commands.sync import SyncCommand


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('-V', '--version', action='store_true', help='Display the version of the Spotty')

    command_classes = [
       StartCommand,
       StopCommand,
       StatusCommand,
       ShCommand,
       RunCommand,
       ExecCommand,
       SyncCommand,
       DownloadCommand,
       AwsCommand,
    ] + _get_custom_commands()

    # add commands to the parser
    add_subparsers(parser, command_classes)

    return parser


def add_subparsers(parser: argparse.ArgumentParser, command_classes: List[Type[AbstractCommand]]):
    """Adds commands to the parser."""
    subparsers = parser.add_subparsers()
    for command_class in command_classes:
        command = command_class()
        subparser = subparsers.add_parser(command.name, help=command.description, description=command.description)
        subparser.set_defaults(command=command, parser=subparser)
        command.configure(subparser)


def _get_custom_commands() -> List[Type[AbstractCommand]]:
    """Returns custom commands that integrated through entry points."""
    return [entry_point.load() for entry_point in pkg_resources.iter_entry_points('spotty.commands')]



from argparse import ArgumentParser, Namespace
from spotty.commands.abstract_config_command import AbstractConfigCommand
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.deployment.utils.commands import get_script_command, get_log_command, get_tmux_session_command, get_bash_command
from spotty.errors.instance_not_running import InstanceNotRunningError
from spotty.errors.nothing_to_do import NothingToDoError
from spotty.deployment.utils.user_scripts import parse_script_parameters, render_script
from spotty.deployment.abstract_instance_manager import AbstractInstanceManager


class RunCommand(AbstractConfigCommand):

    name = 'run'
    description = 'Run a custom script from the configuration file inside the container'

    def configure(self, parser: ArgumentParser):
        super().configure(parser)
        parser.add_argument('script_name', metavar='SCRIPT_NAME', type=str, help='Script name')
        parser.add_argument('-u', '--user', type=str, default=None,
                            help='Container username or UID (format: <name|uid>[:<group|gid>])')
        parser.add_argument('-s', '--session-name', type=str, default=None, help='tmux session name')
        parser.add_argument('-l', '--logging', action='store_true', help='Log the script outputs to a file')
        parser.add_argument('-p', '--parameter', metavar='PARAMETER=VALUE', action='append', type=str, default=[],
                            help='Set a value for the script parameter (format: PARAMETER=VALUE). This '
                                 'argument can be used multiple times to set several parameters. Parameters can be '
                                 'used in the script as Mustache variables (for example: {{PARAMETER}}).')
        parser.add_argument('--no-sync', action='store_true', help='Don\'t sync the project before running the script')

        # add the "double-dash" argument to the usage message
        parser.prog = 'spotty run'
        parser.usage = parser.format_usage()[7:-1] + ' [-- args...]\n'
        parser.epilog = 'The double dash (--) separates custom arguments that you can pass to the script ' \
                        'from the Spotty arguments.'

    def _run(self, instance_manager: AbstractInstanceManager, args: Namespace, output: AbstractOutputWriter):
        # check that the script exists
        script_name = args.script_name
        scripts = instance_manager.project_config.scripts
        if script_name not in scripts:
            raise ValueError('Script "%s" is not defined in the configuration file.' % script_name)

        # replace script parameters
        params = parse_script_parameters(args.parameter)
        script_content = render_script(scripts[script_name], params)

        # check that the instance is started
        if not instance_manager.is_running():
            raise InstanceNotRunningError(instance_manager.instance_config.name)

        # sync the project with the instance
        if not args.no_sync:
            try:
                instance_manager.sync(output)
            except NothingToDoError:
                pass

        # get a command to run the script with "docker exec"
        script_command = get_script_command(script_name, script_content, script_args=args.custom_args,
                                            logging=args.logging)
        command = instance_manager.container_commands.exec(script_command, interactive=True, tty=True,
                                                           user=args.user)

        # wrap the command with the tmux session
        if instance_manager.use_tmux:
            session_name = args.session_name if args.session_name else 'spotty-script-%s' % script_name
            default_command = instance_manager.container_commands.exec(get_bash_command(), interactive=True, tty=True,
                                                                       user=args.user)
            command = get_tmux_session_command(command, session_name, script_name, default_command=default_command,
                                               keep_pane=True)

        # execute command on the host OS
        instance_manager.exec(command)



import os
import subprocess
import tempfile
from glob import glob
from spotty.providers.aws.helpers.s3_sync import get_s3_sync_command


def get_logs_s3_path(bucket_name: str, instance_name: str) -> str:
    return 's3://%s/logs/aws/%s' % (bucket_name, instance_name)


def download_logs(bucket_name: str, instance_name: str, stack_uuid: str, region: str) -> list:
    """Downloads logs from S3 bucket to temporary directory."""
    logs_s3_path = '%s/%s' % (get_logs_s3_path(bucket_name, instance_name), stack_uuid)
    local_logs_dir = tempfile.mkdtemp()

    # download logs
    download_cmd = get_s3_sync_command(logs_s3_path, local_logs_dir, region=region, exact_timestamp=True, quiet=True)
    subprocess.call(download_cmd, shell=True)

    # get paths to the downloaded files
    log_paths = glob(os.path.join(local_logs_dir, '**', '*'), recursive=True)

    return log_paths



from spotty.errors.instance_not_running import InstanceNotRunningError
from spotty.deployment.abstract_cloud_instance.abstract_cloud_instance_manager import AbstractCloudInstanceManager
from spotty.providers.aws.resource_managers.bucket_manager import BucketManager
from spotty.providers.aws.config.instance_config import InstanceConfig
from spotty.providers.aws.data_transfer import DataTransfer
from spotty.providers.aws.instance_deployment import InstanceDeployment
from spotty.utils import render_table


class InstanceManager(AbstractCloudInstanceManager):

    instance_config: InstanceConfig
    bucket_manager: BucketManager
    data_transfer: DataTransfer
    instance_deployment: InstanceDeployment

    def _get_instance_config(self, instance_config: dict) -> InstanceConfig:
        """Validates the instance config and returns an InstanceConfig object."""
        return InstanceConfig(instance_config, self.project_config)

    def _get_bucket_manager(self) -> BucketManager:
        """Returns an bucket manager."""
        return BucketManager(self.instance_config.project_config.project_name, self.instance_config.region)

    def _get_data_transfer(self) -> DataTransfer:
        """Returns a data transfer object."""
        return DataTransfer(
            local_project_dir=self.project_config.project_dir,
            host_project_dir=self.instance_config.host_project_dir,
            sync_filters=self.project_config.sync_filters,
            instance_name=self.instance_config.name,
            region=self.instance_config.region,
        )

    def _get_instance_deployment(self) -> InstanceDeployment:
        """Returns an instance deployment manager."""
        return InstanceDeployment(self.instance_config)

    def get_status_text(self):
        instance = self.instance_deployment.get_instance()
        if not instance:
            raise InstanceNotRunningError(self.instance_config.name)

        table = [
            ('Instance State', instance.state),
            ('Instance Type', instance.instance_type),
            ('Availability Zone', instance.availability_zone),
        ]

        if instance.public_ip_address:
            table.append(('Public IP Address', instance.public_ip_address))

        if instance.lifecycle == 'spot':
            spot_price = instance.get_spot_price()
            table.append(('Purchasing Option', 'Spot Instance'))
            table.append(('Spot Instance Price', '$%.04f' % spot_price))
        else:
            on_demand_price = instance.get_on_demand_price()
            table.append(('Purchasing Option', 'On-Demand Instance'))
            table.append(('Instance Price', ('$%.04f (us-east-1)' % on_demand_price) if on_demand_price else 'Unknown'))

        return render_table(table)

    @property
    def ssh_key_path(self):
        return self.instance_deployment.key_pair_manager.key_path



import os
from typing import List
from spotty.config.abstract_instance_config import AbstractInstanceConfig
from spotty.config.abstract_instance_volume import AbstractInstanceVolume
from spotty.config.project_config import ProjectConfig
from spotty.config.host_path_volume import HostPathVolume
from spotty.providers.remote.config.validation import validate_instance_parameters


class InstanceConfig(AbstractInstanceConfig):

    def __init__(self, instance_config: dict, project_config: ProjectConfig):
        super().__init__(instance_config, project_config)

    def _validate_instance_params(self, params: dict):
        # validate the config and fill missing parameters with the default values
        return validate_instance_parameters(params)

    @property
    def user(self) -> str:
        return self._params['user']

    @property
    def host(self) -> str:
        return self._params['host']

    @property
    def port(self) -> int:
        return self._params['port']

    @property
    def key_path(self) -> str:
        key_path = os.path.expanduser(self._params['keyPath'])
        if not os.path.isabs(key_path):
            key_path = os.path.join(self.project_config.project_dir, key_path)

        key_path = os.path.normpath(key_path)

        return key_path

    def _get_instance_volumes(self) -> List[AbstractInstanceVolume]:
        volumes = []
        for volume_config in self._params['volumes']:
            volume_type = volume_config['type']
            if volume_type == HostPathVolume.TYPE_NAME:
                volumes.append(HostPathVolume(volume_config))
            else:
                raise ValueError('Volume type "%s" is not supported.' % volume_type)

        return volumes



from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.errors.instance_not_running import InstanceNotRunningError
from spotty.deployment.abstract_cloud_instance.abstract_cloud_instance_manager import AbstractCloudInstanceManager
from spotty.providers.gcp.config.instance_config import InstanceConfig
from spotty.providers.gcp.data_transfer import DataTransfer
from spotty.providers.gcp.instance_deployment import InstanceDeployment
from spotty.providers.gcp.resource_managers.bucket_manager import BucketManager
from spotty.utils import render_table


class InstanceManager(AbstractCloudInstanceManager):

    instance_config: InstanceConfig
    bucket_manager: BucketManager
    data_transfer: DataTransfer
    instance_deployment: InstanceDeployment

    def _get_instance_config(self, instance_config: dict) -> InstanceConfig:
        """Validates the instance config and returns an InstanceConfig object."""
        return InstanceConfig(instance_config, self.project_config)

    def _get_bucket_manager(self) -> BucketManager:
        region = '-'.join(self.instance_config.zone.split('-')[:-1])
        return BucketManager(self.instance_config.project_config.project_name, region)

    def _get_data_transfer(self) -> DataTransfer:
        """Returns a data transfer object."""
        return DataTransfer(
            local_project_dir=self.project_config.project_dir,
            host_project_dir=self.instance_config.host_project_dir,
            sync_filters=self.project_config.sync_filters,
            instance_name=self.instance_config.name,
        )

    def _get_instance_deployment(self) -> InstanceDeployment:
        """Returns an instance deployment manager."""
        return InstanceDeployment(self.instance_config)

    def download(self, download_filters: list, output: AbstractOutputWriter, dry_run=False):
        raise NotImplementedError('GCP provider doesn\'t have an implementation of the "download" command yet.')

    def get_status_text(self) -> str:
        instance = self.instance_deployment.get_instance()
        if not instance:
            raise InstanceNotRunningError(self.instance_config.name)

        table = [
            ('Instance Status', instance.status),
            ('Machine Type', instance.machine_type),
            ('Zone', instance.zone),
        ]

        if instance.public_ip_address:
            table.append(('Public IP Address', instance.public_ip_address))

        table.append(('Purchasing Option', 'Preemtible VM' if instance.is_preemtible else 'On-demand VM'))

        return render_table(table)

    @property
    def ssh_key_path(self):
        return self.instance_deployment.ssh_key_manager.private_key_file



from spotty.providers.gcp.helpers.ce_client import CEClient


class Disk(object):

    def __init__(self, ce: CEClient, data: dict):
        """
        Args:
            data (dict): Example:
               {'creationTimestamp': '2019-04-20T16:21:49.579-07:00',
                'guestOsFeatures': [{'type': 'VIRTIO_SCSI_MULTIQUEUE'}],
                'id': '1546539587132069731',
                'kind': 'compute#disk',
                'labelFingerprint': '42WmSpB8rSM=',
                'lastAttachTimestamp': '2019-04-20T16:21:49.580-07:00',
                'licenseCodes': ['1000205'],
                'licenses': ['https://www.googleapis.com/compute/v1/projects/debian-cloud/global/licenses/debian-9-stretch'],
                'name': 'instance-1',
                'physicalBlockSizeBytes': '4096',
                'selfLink': 'https://www.googleapis.com/compute/v1/projects/spotty-221422/zones/us-east1-b/disks/instance-1',
                'sizeGb': '10',
                'sourceImage': 'https://www.googleapis.com/compute/v1/projects/debian-cloud/global/images/debian-9-stretch-v20190326',
                'sourceImageId': '6831652533131678657',
                'status': 'READY',
                'type': 'https://www.googleapis.com/compute/v1/projects/spotty-221422/zones/us-east1-b/diskTypes/pd-standard',
                'users': ['https://www.googleapis.com/compute/v1/projects/spotty-221422/zones/us-east1-b/instances/instance-1'],
                'zone': 'https://www.googleapis.com/compute/v1/projects/spotty-221422/zones/us-east1-b'}
        """
        self._ce = ce
        self._data = data

    @staticmethod
    def get_by_name(ce: CEClient, disk_name: str):
        """Returns a disk by its name."""
        res = ce.list_disks(disk_name)
        if not res:
            return None

        return Disk(ce, res[0])

    @property
    def name(self) -> str:
        return self._data['name']

    @property
    def status(self) -> str:
        return self._data['status']

    @property
    def size(self) -> int:
        return int(self._data['sizeGb'])

    @property
    def users(self) -> list:
        return self._data.get('users', [])

    def is_available(self):
        return (self.status == 'READY') and not self.users



import logging
from collections import OrderedDict
from time import sleep
from httplib2 import ServerNotFoundError
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.gcp.resources.instance import Instance
from spotty.providers.gcp.resources.stack import Stack
from spotty.providers.gcp.helpers.ce_client import CEClient
from spotty.providers.gcp.helpers.dm_client import DMClient
from spotty.providers.gcp.helpers.dm_resource import DMResource


def wait_resources(dm: DMClient, ce: CEClient, deployment_name: str, resource_messages: OrderedDict,
                   instance_resource_name: str, machine_name: str, output: AbstractOutputWriter, delay: int = 5):
    # make sure that the instance resource is in the messages list
    assert any(resource_name == instance_resource_name for resource_name, _ in resource_messages.items())

    created_resources = set()
    for resource_name, message in resource_messages.items():
        output.write('- %s...' % message)

        is_created = False
        while not is_created:
            sleep(delay)

            # get the resource info
            try:
                # check that the deployment is not failed
                stack = Stack.get_by_name(dm, deployment_name)
                if stack.error:
                    raise ValueError('Deployment "%s" failed.\n'
                                     'Error: %s' % (deployment_name, stack.error['message']))

                # check if the instance was preempted, terminated or deleted right after creation
                if instance_resource_name in created_resources:
                    instance = Instance.get_by_name(ce, machine_name)
                    if not instance or instance.is_stopped:
                        raise ValueError('Error: the instance was unexpectedly terminated. Please, check out the '
                                         'instance logs to find out the reason.\n')

                # get resource
                resource = DMResource.get_by_name(dm, deployment_name, resource_name)
            except (ConnectionResetError, ServerNotFoundError):
                logging.warning('Connection problem')
                continue

            # resource doesn't exist yet
            if not resource:
                continue

            # resource failed
            if resource.is_failed:
                error_msg = ('Error: ' + resource.error_message) if resource.error_message \
                    else 'Please, see Deployment Manager logs for the details.' % deployment_name

                raise ValueError('Deployment "%s" failed.\n%s' % (deployment_name, error_msg))

            # resource was successfully created
            is_created = resource.is_created

        created_resources.add(resource_name)


def check_gpu_configuration(ce: CEClient, gpu_parameters: dict):
    if not gpu_parameters:
        return

    # check GPU type
    accelerator_types = ce.get_accelerator_types()
    gpu_type = gpu_parameters['type']
    if gpu_type not in accelerator_types:
        if accelerator_types:
            error_msg = 'GPU type "%s" is not supported in the "%s" zone.\nAvailable GPU types are: %s.' \
                        % (gpu_type, ce.zone, ', '.join(accelerator_types.keys()))
        else:
            error_msg = 'The "%s" zone doesn\'t support any GPU accelerators.' % ce.zone

        raise ValueError(error_msg)

    # check the number of GPUs is not exceed the maximum
    max_cards_per_instance = accelerator_types[gpu_parameters['type']]
    if gpu_parameters['count'] > max_cards_per_instance:
        raise ValueError('Maximum allowed number of cards per instance for the "%s" type is %d.'
                         % (gpu_parameters['type'], max_cards_per_instance))




from collections import namedtuple
from time import sleep
from typing import List, Dict
from botocore.exceptions import EndpointConnectionError, ClientError
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter


Task = namedtuple('Task', ['message', 'start_resource', 'finish_resource', 'enabled'])


class Stack(object):

    def __init__(self, cf, stack_info):
        self._cf = cf
        self._stack_info = stack_info

    @staticmethod
    def get_by_name(cf, stack_name: str):
        """Returns a Stack by its name."""
        try:
            res = cf.describe_stacks(StackName=stack_name)
        except ClientError as e:
            # ignore an exception if it raised because the stack doesn't exist
            error_code = e.response.get('Error', {}).get('Code')
            if error_code != 'ValidationError':
                raise e

            res = {'Stacks': []}

        if not len(res['Stacks']):
            return None

        return Stack(cf, res['Stacks'][0])

    @staticmethod
    def create_stack(cf, *args, **kwargs):
        res = cf.create_stack(*args, **kwargs)
        return Stack(cf, res)

    @staticmethod
    def update_stack(cf, *args, **kwargs):
        res = cf.update_stack(*args, **kwargs)
        return Stack(cf, res)

    @property
    def stack_id(self) -> str:
        return self._stack_info['StackId']

    @property
    def stack_uuid(self) -> str:
        return self.stack_id.rsplit('/', 1)[-1]

    @property
    def name(self) -> str:
        return self._stack_info['StackName']

    @property
    def status(self) -> str:
        return self._stack_info['StackStatus']

    @property
    def outputs(self) -> str:
        return self._stack_info['Outputs']

    def delete(self):
        return self._cf.delete_stack(StackName=self.stack_id)

    def wait_stack_created(self, delay=30):
        waiter = self._cf.get_waiter('stack_create_complete')
        waiter.wait(StackName=self.stack_id, WaiterConfig={'Delay': delay})

    def wait_stack_updated(self, delay=30):
        waiter = self._cf.get_waiter('stack_update_complete')
        waiter.wait(StackName=self.stack_id, WaiterConfig={'Delay': delay})

    def wait_stack_deleted(self, delay=30):
        waiter = self._cf.get_waiter('stack_delete_complete')
        waiter.wait(StackName=self.stack_id, WaiterConfig={'Delay': delay})

    def wait_status_changed(self, stack_waiting_status: str, output: AbstractOutputWriter, delay: int = 5):
        stack = None
        while True:
            # get the latest status of the stack
            try:
                stack = self.get_by_name(self._cf, self.stack_id)
            except EndpointConnectionError as e:
                output.write(str(e))
                continue

            if stack.status != stack_waiting_status:
                break

            sleep(delay)

        return stack

    def wait_tasks(self, tasks: List[Task], resource_success_status: str, resource_fail_status: str,
                   output: AbstractOutputWriter, delay: int = 5):
        resource_statuses = self._get_resource_statuses()

        for task in tasks:
            if not task.enabled:
                continue

            task_started = task_finished = False
            while not task_finished:
                start_status = resource_statuses.get(task.start_resource)
                finish_status = resource_statuses.get(task.finish_resource)

                if not task_started and (not task.start_resource or (start_status == resource_success_status)):
                    task_started = True
                    output.write('- %s... ' % task.message, newline=False)
                elif task_started and (finish_status == resource_success_status):
                    task_finished = True
                    output.write('DONE')
                else:
                    sleep(delay)
                    try:
                        resource_statuses = self._get_resource_statuses()
                    except EndpointConnectionError as e:
                        output.write(str(e))
                        continue

                    # check that the stack is not failed
                    for status in resource_statuses.values():
                        if status == resource_fail_status:
                            if task_started and not task_finished:
                                output.write('')
                            return

    def _get_resource_statuses(self) -> Dict[str, str]:
        stack_resources = self._cf.list_stack_resources(StackName=self.stack_id)

        resource_statuses = {row['LogicalResourceId']: row['ResourceStatus']
                             for row in stack_resources['StackResourceSummaries']}

        return resource_statuses



from spotty.providers.gcp.config.instance_config import DEFAULT_IMAGE_NAME
from spotty.providers.gcp.helpers.ce_client import CEClient
from spotty.providers.gcp.resources.image import Image


def get_image(ce: CEClient, image_uri: str = None, image_name: str = None) -> Image:
    """Returns an image that should be used for deployment.

    Raises:
        ValueError: If an image not found.
    """
    if image_uri:
        # get an image by its URL if the "imageUri" parameter is specified
        image = Image.get_by_uri(ce, image_uri)
        if not image:
            raise ValueError('Image "%s" not found.' % image_uri)
    elif image_name:
        # get an image by name if the "imageName" parameter is specified
        image = Image.get_by_name(ce, image_name)
        if not image:
            # if an image name was explicitly specified, but the image was not found, raise an error
            raise ValueError('Image with the name "%s" was not found.' % image_name)
    else:
        # if the "imageName" parameter is not specified, try to use the default image name
        image = Image.get_by_name(ce, DEFAULT_IMAGE_NAME)
        if not image:
            # get the latest "common-gce-gpu-image" image
            image_family_url = 'projects/ml-images/global/images/family/common-gce-gpu-image'
            image = Image.get_by_uri(ce, image_family_url)
            if not image:
                raise ValueError('The "common-gce-gpu-image" image was not found.')

    return image



import re


IMAGE_URI_REGEX = '^(?:(?:https://compute.googleapis.com/compute/v1/)?projects/([a-z](?:[-a-z0-9]*[a-z0-9])?)/)?' \
                  'global/images/(family/)?([a-z](?:[-a-z0-9]*[a-z0-9])?)$'


class ImageUri(object):

    def __init__(self, image_uri: str):
        res = re.match(IMAGE_URI_REGEX, image_uri)
        if not res:
            raise ValueError('Image URI has a wrong format')

        self._project_id, self._is_family, self._name = res.groups()

    @property
    def project_id(self) -> str:
        return self._project_id

    @property
    def is_family(self):
        return bool(self._is_family)

    @property
    def name(self):
        """Image name or image family name."""
        return self._name



import logging
from time import sleep
from httplib2 import ServerNotFoundError
from spotty.providers.gcp.helpers.dm_client import DMClient


class Stack(object):

    def __init__(self, dm: DMClient, data: dict):
        """
        Args:
            dm (DMClient): Deployment Manager client
            data (dict): Stack info. Example:
               {'fingerprint': 'vvtbwT7F953T0YC9tQ9CUg==',
                'id': '3128259442476717093',
                'insertTime': '2019-04-20T16:27:06.739-07:00',
                'name': 'spotty-instance-x11-test-i2',
                'operation': {'endTime': '2019-04-20T16:27:23.141-07:00',
                              'error': {'errors': [{'code': 'RESOURCE_ERROR',
                                                    'location': '/deployments/spotty-instance-x11-test-i2/resources/spotty-instance-x11-test-i2-disk-1',
                                                    'message': '{"ResourceType":"compute.v1.disk","ResourceErrorCode":"400","ResourceErrorMessage":{"code":400,"errors":[{"domain":"global","location":"zone","locationType":"parameter","message":"Invalid '
                                                               "value 'zones/us-east1-b'. "
                                                               'Values must match the '
                                                               'following regular expression: '
                                                               '\'[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?\'","reason":"invalidParameter"}],"message":"Invalid '
                                                               "value 'zones/us-east1-b'. "
                                                               'Values must match the '
                                                               'following regular expression: '
                                                               '\'[a-z](?:[-a-z0-9]{0,61}[a-z0-9])?\'","statusMessage":"Bad '
                                                               'Request","requestPath":"https://www.googleapis.com/compute/v1/projects/spotty-221422/zones/zones%2Fus-east1-b/disks","httpMethod":"POST"}}'}]},
                              'httpErrorMessage': 'BAD REQUEST',
                              'httpErrorStatusCode': 400,
                              'id': '1965130840716743717',
                              'insertTime': '2019-04-20T16:27:06.900-07:00',
                              'kind': 'deploymentmanager#operation',
                              'name': 'operation-1555802826515-586fe92d0a605-6166491c-7051043e',
                              'operationType': 'insert',
                              'progress': 100,
                              'selfLink': 'https://www.googleapis.com/deploymentmanager/v2/projects/spotty-221422/global/operations/operation-1555802826515-586fe92d0a605-6166491c-7051043e',
                              'startTime': '2019-04-20T16:27:06.908-07:00',
                              'status': 'DONE',
                              'targetId': '3128259442476717093',
                              'targetLink': 'https://www.googleapis.com/deploymentmanager/v2/projects/spotty-221422/global/deployments/spotty-instance-x11-test-i2',
                              'user': 'spotty@spotty-221422.iam.gserviceaccount.com'},
                'selfLink': 'https://www.googleapis.com/deploymentmanager/v2/projects/spotty-221422/global/deployments/spotty-instance-x11-test-i2',
                'update': {'manifest': 'https://www.googleapis.com/deploymentmanager/v2/projects/spotty-221422/global/deployments/spotty-instance-x11-test-i2/manifests/manifest-1555802826772'},
                'updateTime': '2019-04-20T16:27:23.107-07:00'}
        """
        self._dm = dm
        self._data = data

    @staticmethod
    def get_by_name(dm: DMClient, deployment_name: str):
        """Returns an instance by its stack name."""
        res = dm.get(deployment_name)
        if not res:
            return None

        return Stack(dm, res)

    @staticmethod
    def create(dm: DMClient, deployment_name: str, template: str):
        return dm.deploy(deployment_name, template)

    @property
    def name(self) -> str:
        return self._data['name']

    @property
    def status(self) -> str:
        return self._data.get('operation', {}).get('status')

    @property
    def is_running(self):
        return self.status == 'RUNNING'

    @property
    def is_done(self):
        """A deployment has the done status when it's successfully created or failed."""
        return self.status == 'DONE'

    @property
    def error(self) -> str:
        """Returns an error in the format: {'code': '...', 'message': '...'}."""
        return self._data.get('operation', {}).get('error', {}).get('errors', [None])[0]

    @property
    def fingerprint(self) -> str:
        return self._data['fingerprint']

    def stop(self):
        self._dm.stop(self.name, self.fingerprint)

    def delete(self):
        self._dm.delete(self.name)

    def wait_stack_deleted(self, delay=15):
        stack = True
        while stack:
            try:
                stack = self.get_by_name(self._dm, self.name)
            except (ConnectionResetError, ServerNotFoundError):
                logging.warning('Connection problem')
                continue

            sleep(delay)

    def wait_stack_done(self, delay=5):
        is_done = False
        while not is_done:
            try:
                stack = self.get_by_name(self._dm, self.name)
                is_done = stack.is_done
            except (ConnectionResetError, ServerNotFoundError):
                logging.warning('Connection problem')
                continue

            sleep(delay)



from abc import abstractmethod, ABC
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.config.abstract_instance_config import AbstractInstanceConfig
from spotty.deployment.abstract_cloud_instance.abstract_data_transfer import AbstractDataTransfer
from spotty.deployment.abstract_cloud_instance.resources.abstract_instance import AbstractInstance
from spotty.deployment.container.abstract_container_commands import AbstractContainerCommands


class AbstractInstanceDeployment(ABC):

    def __init__(self, instance_config: AbstractInstanceConfig):
        self._instance_config = instance_config

    @property
    def instance_config(self) -> AbstractInstanceConfig:
        return self._instance_config

    @abstractmethod
    def get_instance(self) -> AbstractInstance:
        """Returns information about the instance it it exists."""
        raise NotImplementedError

    @abstractmethod
    def deploy(self, container_commands: AbstractContainerCommands, bucket_name: str,
               data_transfer: AbstractDataTransfer, output: AbstractOutputWriter, dry_run: bool = False):
        """Deploys or redeploys the instance."""
        raise NotImplementedError

    @abstractmethod
    def delete(self, output: AbstractOutputWriter):
        """Deletes the stack with the instance and applies deletion policies for the volumes."""
        raise NotImplementedError



import boto3
import re
from spotty.deployment.abstract_cloud_instance.abstract_bucket_manager import AbstractBucketManager
from spotty.deployment.abstract_cloud_instance.errors.bucket_not_found import BucketNotFoundError
from spotty.providers.aws.resources.bucket import Bucket
from spotty.utils import random_string


class BucketManager(AbstractBucketManager):

    def __init__(self, project_name: str, region: str):
        super().__init__(project_name)

        self._s3 = boto3.client('s3', region_name=region)
        self._region = region
        self._bucket_prefix = 'spotty-%s' % project_name.lower()

    def get_bucket(self) -> Bucket:
        res = self._s3.list_buckets()
        regex = re.compile('-'.join([self._bucket_prefix, '[a-z0-9]{12}', self._region]))
        buckets = [bucket for bucket in res['Buckets'] if regex.match(bucket['Name']) is not None]

        if len(buckets) > 1:
            raise ValueError('Found several buckets in the same region: %s.'
                             % ', '.join(bucket['Name'] for bucket in buckets))

        if not len(buckets):
            raise BucketNotFoundError

        bucket = Bucket(buckets[0])

        return bucket

    def create_bucket(self) -> Bucket:
        bucket_name = '-'.join([self._bucket_prefix, random_string(12), self._region])

        # a fix for the boto3 issue: https://github.com/boto/boto3/issues/125
        if self._region == 'us-east-1':
            self._s3.create_bucket(ACL='private', Bucket=bucket_name)
        else:
            self._s3.create_bucket(ACL='private', Bucket=bucket_name,
                                   CreateBucketConfiguration={'LocationConstraint': self._region})

        return Bucket({'Name': bucket_name})

    def delete_bucket(self):
        pass



from collections import OrderedDict
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.providers.gcp.resources.stack import Stack
from spotty.providers.gcp.helpers.ce_client import CEClient
from spotty.providers.gcp.helpers.deployment import wait_resources
from spotty.providers.gcp.helpers.dm_client import DMClient
from spotty.providers.gcp.helpers.dm_resource import DMResource
from spotty.providers.gcp.helpers.rtc_client import RtcClient


class InstanceStackManager(object):

    def __init__(self, machine_name: str, project_id: str, zone: str):
        self._dm = DMClient(project_id, zone)
        self._ce = CEClient(project_id, zone)
        self._rtc = RtcClient(project_id, zone)
        self._machine_name = machine_name
        self._stack_name = 'spotty-instance-' + machine_name

        # resource names
        self._INSTANCE_RESOURCE_NAME = machine_name
        self._DOCKER_WAITER_RESOURCE_NAME = machine_name + '-docker-waiter'
        self._DOCKER_STATUS_CONFIG_RESOURCE_NAME = machine_name + '-docker-status'

    @property
    def name(self):
        return self._stack_name

    def create_stack(self, template: str, output: AbstractOutputWriter):
        """Deploys a Deployment Manager template."""

        # create a stack
        res = Stack.create(self._dm, self._stack_name, template)
        # print(res)
        # exit()

        output.write('Waiting for the stack to be created...')

        resource_messages = OrderedDict([
            (self._INSTANCE_RESOURCE_NAME, 'launching the instance'),
            (self._DOCKER_WAITER_RESOURCE_NAME, 'running the Docker container'),
        ])

        # wait for the stack to be created
        with output.prefix('  '):
            wait_resources(self._dm, self._ce, self._stack_name, resource_messages,
                           instance_resource_name=self._INSTANCE_RESOURCE_NAME, machine_name=self._machine_name,
                           output=output)

    def delete_stack(self, output: AbstractOutputWriter):
        stack = Stack.get_by_name(self._dm, self._stack_name)
        if not stack:
            return

        output.write('Waiting for the stack to be deleted...')

        # delete the stack
        try:
            if stack.is_running:
                # stop an ongoing operation first to make sure the delete method
                # won't raise an error "Resource '...' has an ongoing conflicting operation"
                stack.stop()

                # if the docker-waiter resource is still waiting for a signal, send a failure signal
                # to be able to delete the stack
                resource = DMResource.get_by_name(self._dm, self._stack_name, self._DOCKER_WAITER_RESOURCE_NAME)
                if resource.is_in_progress:
                    self._rtc.set_value(self._DOCKER_STATUS_CONFIG_RESOURCE_NAME, '/failure/1', '1')

                # wait until the stack will be created or will fail
                stack.wait_stack_done()

            stack.delete()
            stack.wait_stack_deleted()
        except Exception as e:
            raise ValueError('Stack "%s" was not deleted. Error: %s\n'
                             'See Deployment Manager logs for details.' % (self._stack_name, str(e)))



from abc import ABC, abstractmethod


class AbstractInstanceVolume(ABC):

    def __init__(self, volume_config: dict):
        self._name = volume_config['name']
        self._params = self._validate_volume_parameters(volume_config['parameters'])

    @abstractmethod
    def _validate_volume_parameters(self, params: dict) -> dict:
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Unique name of the volume that will be used for the deployment."""
        return self._name

    @property
    @abstractmethod
    def host_path(self) -> str:
        """A path on the host OS that will be mounted to the container."""
        raise NotImplementedError

    @property
    @abstractmethod
    def title(self) -> str:
        """A title for the volume type.
        It will be used to display information about the volumes during the deployment.
        """
        raise NotImplementedError

    @property
    def deletion_policy_title(self) -> str:
        """A title for the volume's deletion policy.
        It will be used to display information about the volumes during the deployment.
        """
        return ''



from abc import ABC, abstractmethod
from spotty.deployment.container.abstract_container_commands import AbstractContainerCommands


class AbstractContainerScript(ABC):

    def __init__(self, container_commands: AbstractContainerCommands):
        self._commands = container_commands

    @property
    def commands(self) -> AbstractContainerCommands:
        return self._commands

    @abstractmethod
    def render(self) -> str:
        raise NotImplementedError



from typing import List
from spotty.config.abstract_instance_config import AbstractInstanceConfig, VolumeMount
from spotty.config.abstract_instance_volume import AbstractInstanceVolume
from spotty.providers.aws.config.validation import validate_instance_parameters
from spotty.providers.aws.config.ebs_volume import EbsVolume


DEFAULT_AMI_NAME = 'SpottyAMI'


class InstanceConfig(AbstractInstanceConfig):

    def _validate_instance_params(self, params: dict) -> dict:
        return validate_instance_parameters(params)

    def _get_instance_volumes(self) -> List[AbstractInstanceVolume]:
        volumes = []
        for volume_config in self._params['volumes']:
            volume_type = volume_config['type']
            if volume_type == EbsVolume.TYPE_NAME:
                volumes.append(EbsVolume(volume_config, self.project_config.project_name, self.name))
            else:
                raise ValueError('AWS volume type "%s" not supported.' % volume_type)

        return volumes

    @property
    def user(self):
        return 'ubuntu'

    @property
    def ec2_instance_name(self) -> str:
        return '%s-%s' % (self.project_config.project_name.lower(), self.name.lower())

    @property
    def region(self) -> str:
        return self._params['region']

    @property
    def availability_zone(self) -> str:
        return self._params['availabilityZone']

    @property
    def subnet_id(self) -> str:
        return self._params['subnetId']

    @property
    def instance_type(self) -> str:
        return self._params['instanceType']

    @property
    def is_spot_instance(self) -> bool:
        return self._params['spotInstance']

    @property
    def ami_name(self) -> str:
        return self._params['amiName']

    @property
    def ami_id(self) -> str:
        return self._params['amiId']

    @property
    def root_volume_size(self) -> int:
        return self._params['rootVolumeSize']

    @property
    def ports(self) -> List[int]:
        return list(set(self._params['ports']))

    @property
    def max_price(self) -> float:
        return self._params['maxPrice']

    @property
    def managed_policy_arns(self) -> list:
        return self._params['managedPolicyArns']
    
    @property
    def instance_profile_arn(self) -> str:
        return self._params['instanceProfileArn']



import os
import time
import chevron
from spotty.deployment.utils.commands import get_script_command
from spotty.deployment.container.docker.scripts.abstract_docker_script import AbstractDockerScript


class StartContainerScript(AbstractDockerScript):

    def _partials(self) -> dict:
        return {
            'before_image_build': '',
            'before_container_run': '',
            'before_startup_commands': '',
        }

    def render(self, print_trace: bool = False) -> str:
        # read template file
        template_path = os.path.join(os.path.dirname(__file__), 'data', 'start_container.sh.tpl')
        with open(template_path) as f:
            template = f.read()

        # generate "docker build" command if necessary
        if self.commands.instance_config.dockerfile_path:
            image_name = '%s:%d' % (self.commands.instance_config.full_container_name, time.time())
            build_image_cmd = self.commands.build(image_name)
            pull_image_cmd = ''
        else:
            image_name = self.commands.instance_config.container_config.image
            build_image_cmd = ''
            pull_image_cmd = self.commands.pull()

        # generate a command to run the startup script
        exec_script_cmd = ''
        if self.commands.instance_config.container_config.commands:
            startup_script_cmd = get_script_command('container-startup-commands',
                                                    self.commands.instance_config.container_config.commands)
            exec_script_cmd = self.commands.exec(startup_script_cmd, user='root')

        # render the script
        content = chevron.render(template, data={
            'bash_flags': 'set -xe' if print_trace else 'set -e',
            'is_created_cmd': self.commands.is_created(),
            'remove_cmd': self.commands.remove(),
            'build_image_cmd': build_image_cmd,
            'pull_image_cmd': pull_image_cmd,
            'tmp_container_dir': self.commands.instance_config.host_container_dir,
            'start_container_cmd': self.commands.run(image_name),
            'docker_exec_startup_script_cmd': exec_script_cmd,
        }, partials_dict=self._partials())

        return content



#!/usr/bin/env python

import os
import re
from setuptools import setup, find_packages


def get_version():
    root_dir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(root_dir, 'spotty', '__init__.py')) as f:
        content = f.read()

    version_match = re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', content, re.M)
    if not version_match:
        raise RuntimeError('Unable to find version string.')

    return version_match.group(1)


def get_description():
    readme_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'README.md'))
    with open(readme_path, encoding='utf-8') as f:
        description = f.read()

    return description


setup(name='spotty',
      version=get_version(),
      description='Training deep learning models on AWS and GCP instances',
      url='https://github.com/spotty-cloud/spotty',
      author='Oleg Polosin',
      author_email='apls777@gmail.com',
      license='MIT',
      long_description=get_description(),
      long_description_content_type='text/markdown',
      packages=find_packages(exclude=['tests*']),
      package_data={
          'spotty.deployment.container.docker.scripts': ['data/*', 'data/*/*'],
          'spotty.providers.aws.cfn_templates.instance': ['data/*', 'data/*/*'],
          'spotty.providers.aws.cfn_templates.instance_profile': ['data/*', 'data/*/*'],
          'spotty.providers.gcp.dm_templates.instance': ['data/*', 'data/*/*'],
      },
      scripts=['bin/spotty'],
      install_requires=[
          'boto3>=1.9.0',
          'google-api-python-client>=1.7.8',
          'google-cloud-storage>=1.15.0',
          'cfn_flip',  # to work with CloudFormation templates
          'schema',
          'chevron',
      ],
      tests_require=['moto'],
      test_suite='tests',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Natural Language :: English',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ])



import datetime
import json
import logging
import boto3
from pkg_resources import resource_filename


def get_spot_prices(ec2, instance_type: str):
    """Returns current Spot Instance prices for all availability zones for particular instance type and region.
    AWS region specified implicitly in the "ec2" object.
    """
    tomorrow_date = datetime.datetime.today() + datetime.timedelta(days=1)
    res = ec2.describe_spot_price_history(InstanceTypes=[instance_type],
                                          StartTime=tomorrow_date,
                                          ProductDescriptions=['Linux/UNIX'])

    prices_by_zone = {}
    for row in res['SpotPriceHistory']:
        prices_by_zone[row['AvailabilityZone']] = float(row['SpotPrice'])

    return prices_by_zone


def get_current_spot_price(ec2, instance_type, availability_zone=''):
    """Returns the current Spot price for an availability zone.
    If an availability zone is not specified, returns the minimum price for the region.
    """
    spot_prices = get_spot_prices(ec2, instance_type)
    if availability_zone:
        if availability_zone not in spot_prices:
            raise ValueError('Spot price for the "%s" availability zone not found.' % availability_zone)

        current_price = spot_prices[availability_zone]
    else:
        current_price = min(spot_prices.values())

    return current_price


def get_on_demand_price(instance_type: str, region: str):
    client = boto3.client('pricing', region_name='us-east-1')  # the API available only in "us-east-1"

    try:
        response = client.get_products(
            ServiceCode='AmazonEC2',
            Filters=[
                {'Type': 'TERM_MATCH', 'Field': 'location', 'Value': _get_region_name(region)},
                {'Type': 'TERM_MATCH', 'Field': 'instanceType', 'Value': instance_type},
                {'Type': 'TERM_MATCH', 'Field': 'operatingSystem', 'Value': 'Linux'},
                {'Type': 'TERM_MATCH', 'Field': 'tenancy', 'Value': 'shared'},
                {'Type': 'TERM_MATCH', 'Field': 'capacitystatus', 'Value': 'Used'},
            ],
        )

        prices = json.loads(response['PriceList'][0])['terms']['OnDemand']
        price = float(list(list(prices.values())[0]['priceDimensions'].values())[0]['pricePerUnit']['USD'])
    except Exception as e:
        logging.debug('Couldn\'t find a price for the instance: ' + str(e))
        price = None

    return price


def _get_region_name(region: str):
    endpoint_file = resource_filename('botocore', 'data/endpoints.json')
    try:
        with open(endpoint_file, 'r') as f:
            data = json.load(f)
            region_name = data['partitions'][0]['regions'][region]['description']
    except Exception as e:
        logging.debug('Couldn\'t obtain the region name: ' + str(e))
        region_name = None

    return region_name


def check_max_spot_price(ec2, instance_type: str, is_spot_instance: bool, max_price: float,
                         availability_zone: str = ''):
    """Checks that the specified maximum Spot price is less than the
    current Spot price.

    Args:
        ec2: EC2 client
        instance_type (str): Instance Type
        is_spot_instance (bool): True if it's a spot instance
        max_price (float): requested maximum price for the instance
        availability_zone (str): Availability zone to check. If it's an empty string,
            checks the cheapest AZ.

    Raises:
        ValueError: Current price for the instance is higher than the
            maximum price in the configuration file.
    """
    if is_spot_instance and max_price:
        current_price = get_current_spot_price(ec2, instance_type, availability_zone)
        if current_price > max_price:
            raise ValueError('Current price for the instance (%.04f) is higher than the maximum price in the '
                             'configuration file (%.04f).' % (current_price, max_price))



import os
from typing import List
from schema import Schema, And, Use, Optional, Or, Regex, Hook, SchemaError, SchemaForbiddenKeyError


DEFAULT_CONTAINER_NAME = 'default'


def validate_basic_config(data):

    container = And(
        {
            Optional('name', default=DEFAULT_CONTAINER_NAME): And(str, Regex(r'^[\w-]+$')),
            'projectDir': And(str,
                              And(os.path.isabs,
                                  error='Use an absolute path when specifying the project directory'),
                              Use(lambda x: x.rstrip('/'))
                              ),
            Optional('image', default=''): And(str, len),
            Optional('file', default=''): And(str,  # TODO: a proper regex that the filename is valid
                                              Regex(r'^[\w\.\/@-]*$',
                                                    error='Invalid name for a Dockerfile'),
                                              And(lambda x: not x.endswith('/'),
                                                  error='Invalid name for a Dockerfile'),
                                              And(lambda x: not os.path.isabs(x),
                                                  error='Path to the Dockerfile should be relative to the '
                                                        'project\'s root directory.'),
                                              ),
            Optional('runAsHostUser', default=False): bool,
            Optional('volumeMounts', default=[]): (And(
                [{
                    'name': And(Or(int, str), Use(str), Regex(r'^[\w-]+$')),
                    'mountPath': And(
                        str,
                        And(os.path.isabs, error='Use an absolute path when specifying a mount directory'),
                        Use(lambda x: x.rstrip('/')),
                    ),
                }],
                And(lambda x: is_unique_value(x, 'name'),
                    error='Each volume mount must have a unique name.'),
                And(lambda x: not has_prefix([(volume['mountPath'] + '/') for volume in x]),
                    error='Volume mount paths cannot be prefixes for each other.'),
            )),
            Optional('workingDir', default=''): And(str,
                                                    And(os.path.isabs,
                                                        error='Use an absolute path when specifying a '
                                                              'working directory'),
                                                    Use(lambda x: x.rstrip('/'))
                                                    ),
            Optional('env', default={}): {
                And(str, Regex(r'^[a-zA-Z_]+[a-zA-Z0-9_]*$')): str,
            },
            Optional('hostNetwork', default=False): bool,
            Optional('ports', default=[]): [{
                'containerPort': And(int, lambda x: 0 < x < 65536),
                Optional('hostPort', default=None): And(int, lambda x: 0 < x < 65536),
            }],
            Optional('commands', default=''): str,
            # TODO: allow to use only certain runtime parameters
            Optional('runtimeParameters', default=[]): And([str], Use(lambda x: [p.strip() for p in x])),
        },
        And(lambda x: x['image'] or x['file'], error='Either "image" or "file" should be specified.'),
        And(lambda x: not (x['image'] and x['file']), error='"image" and "file" cannot be specified together.'),
        And(lambda x: not (x['hostNetwork'] and x['ports']),
            error='Published ports and the host network mode cannot be used together.'),
    )

    schema = Schema({
        'project': {
            'name': And(str, Regex(r'^[a-zA-Z0-9][a-zA-Z0-9-]{,26}[a-zA-Z0-9]$')),
            Optional('syncFilters', default=[]): And(
                [And(
                    {
                        Optional('exclude'): [And(str, len, And(lambda x: '**' not in x,
                                                                error='Use single asterisks ("*") in sync filters'))],
                        Optional('include'): [And(str, len, And(lambda x: '**' not in x,
                                                                error='Use single asterisks ("*") in sync filters'))],
                    },
                    And(lambda x: x, error='Either "exclude" or "include" filter should be specified.'),
                    And(lambda x: not ('exclude' in x and 'include' in x),
                        error='"exclude" and "include" filters should be specified as different list items.'),
                )],
                error='"project.syncFilters" field must be a list.',
            )
        },
        WrongKey('container', error='Use "containers" field instead of "container".'): object,
        Optional('containers', default=[]): And(
            [container],
            And(lambda x: is_unique_value(x, 'name'), error='Each container must have a unique name.'),
            error='"containers" field must be a list.',
        ),
        WrongKey('instance', error='Use "instances" field instead of "instance".'): object,
        'instances': And(
            [{
                'name': And(Or(int, str), Use(str), Regex(r'^[\w-]+$')),
                'provider': str,
                Optional('parameters', default={}): {
                    And(str, Regex(r'^[\w]+$')): object,
                }
            }],
            And(lambda x: len(x), error='At least one instance must be specified in the configuration file.'),
            And(lambda x: is_unique_value(x, 'name'), error='Each instance must have a unique name.'),
        ),
        Optional('scripts', default={}): {
            And(str, Regex(r'^[\w-]+$')): And(str, len),
        },
    })

    return validate_config(schema, data)


def validate_host_path_volume_parameters(params: dict):
    schema = Schema({
        'path': And(str, Use(lambda x: x.rstrip('/'))),
    })

    return validate_config(schema, params)


def get_instance_parameters_schema(instance_parameters: dict, default_volume_type: str,
                                   instance_checks: list = None, volumes_checks: list = None):
    if not instance_checks:
        instance_checks = []

    if not volumes_checks:
        volumes_checks = []

    schema = Schema(And(
        {
            **instance_parameters,
            Optional('containerName', default=None): And(str, Regex(r'^[\w-]+$')),
            Optional('dockerDataRoot', default=''): And(
                str,
                And(os.path.isabs, error='Use an absolute path when specifying a Docker data root directory'),
                Use(lambda x: x.rstrip('/')),
            ),
            Optional('volumes', default=[]): And(
                [{
                    'name': And(Or(int, str), Use(str), Regex(r'^[\w-]+$')),
                    Optional('type', default=default_volume_type): str,
                    Optional('parameters', default={}): {
                        And(str, Regex(r'^[\w]+$')): object,
                    },
                }],
                And(lambda x: is_unique_value(x, 'name'), error='Each instance volume must have a unique name.'),
                *volumes_checks,
            ),
            Optional('localSshPort', default=None): Or(None, And(int, lambda x: 0 < x < 65536)),
            Optional('commands', default=''): str,
        },
        And(lambda x: not x['dockerDataRoot'] or any([True for v in x['volumes'] if v['parameters'].get('mountDir') and
                                                      is_subdir(x['dockerDataRoot'], v['parameters']['mountDir'])]),
            error='The "mountDir" of one of the volumes must be a prefix for the "dockerDataRoot" path.'),
        *instance_checks
    ))

    return schema


def is_unique_value(x: List[dict], key):
    """Returns "True" if all values of the key in the list of dictionaries are unique."""
    return len(x) == len(set([v[key] for v in x]))


def has_prefix(x: list):
    """Returns "True" if some value in the list is a prefix for another value in this list."""
    for val in x:
        if len(list(filter(val.startswith, x))) > 1:
            return True

    return False


def is_subdir(subdir_path, dir_path):
    """Returns "True" if it's the second path parameter is a subdirectory of the first path parameter."""
    return (subdir_path.rstrip('/') + '/').startswith(dir_path.rstrip('/') + '/')


def validate_config(schema: Schema, config):
    try:
        validated = schema.validate(config)
    except SchemaError as e:
        raise ValueError('Validation error: ' + (e.errors[-1] if e.errors[-1] else e.autos[-1]))

    return validated


class WrongKey(Hook):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, handler=self.raise_error)

    def raise_error(self, key, *args):
        raise SchemaForbiddenKeyError(self._error)



import shlex


def shlex_join(split_command: list):
    """Return a shell-escaped string from *split_command*.
    Copy-pasted from the Python 3.8 code.
    """
    return ' '.join(shlex.quote(arg) for arg in split_command)



import os
from spotty.config.abstract_instance_volume import AbstractInstanceVolume
from spotty.config.validation import validate_host_path_volume_parameters


class HostPathVolume(AbstractInstanceVolume):

    TYPE_NAME = 'HostPath'

    def __init__(self, volume_config: dict, base_dir: str = None):
        super().__init__(volume_config)

        self._base_dir = base_dir

    def _validate_volume_parameters(self, params: dict) -> dict:
        return validate_host_path_volume_parameters(params)

    @property
    def title(self):
        return 'HostPath volume'

    @property
    def name(self):
        return self._name

    @property
    def deletion_policy_title(self) -> str:
        return ''

    @property
    def host_path(self) -> str:
        """A path on the host OS that will be mounted to the container."""
        path = os.path.expanduser(self._params['path'])
        if not os.path.isabs(path):
            if self._base_dir is not None:
                path = os.path.join(self._base_dir, path)
            else:
                raise ValueError('Use absolute path for the "%s" volume.' % self.name)

        return path



from google.auth import default


class GcpCredentials(object):
    def __init__(self):
        credentials, effective_project_id = default()

        self._credentials = credentials
        self._project_id = effective_project_id

    @property
    def project_id(self):
        return self._project_id

    @property
    def service_account_email(self):
        return self._credentials.service_account_email



from spotty.providers.aws.config.instance_config import DEFAULT_AMI_NAME
from spotty.providers.aws.resources.image import Image


def get_ami(ec2, ami_id: str = None, ami_name: str = None) -> Image:
    """Returns an AMI that should be used for deployment.

    Raises:
        ValueError: If an AMI not found.
    """
    if ami_id:
        # get an AMI by ID if the "amiId" parameter is specified
        image = Image.get_by_id(ec2, ami_id)
        if not image:
            raise ValueError('AMI with ID=%s not found.' % ami_id)
    elif ami_name:
        # get an AMI by name if the "amiName" parameter is specified
        image = Image.get_by_name(ec2, ami_name)
        if not image:
            # if an AMI name was explicitly specified in the config, but the AMI was not found, raise an error
            raise ValueError('AMI with the name "%s" was not found.' % ami_name)
    else:
        # if the "amiName" parameter is not specified, try to use the default AMI name
        image = Image.get_by_name(ec2, DEFAULT_AMI_NAME)
        if not image:
            # get the latest "Deep Learning Base AMI"
            res = ec2.describe_images(
                Owners=['amazon'],
                Filters=[{'Name': 'name', 'Values': ['Deep Learning AMI (Ubuntu 16.04) Version*']}],
            )

            if not len(res['Images']):
                raise ValueError('AWS Deep Learning AMI not found.\n'
                                 'Use the "spotty aws create-ami" command to create an AMI with NVIDIA Docker.')

            image_info = sorted(res['Images'], key=lambda x: x['CreationDate'], reverse=True)[0]
            image = Image(ec2, image_info)

    return image



from spotty.providers.gcp.helpers.ce_client import CEClient


class Snapshot(object):

    def __init__(self, data: dict):
        """
        Args:
            data (dict): Example:
               {'creationTimestamp': '2019-04-20T12:40:13.291-07:00',
                'diskSizeGb': '10',
                'id': '714587297862306675',
                'kind': 'compute#snapshot',
                'labelFingerprint': '42WmSpB8rSM=',
                'name': 'snapshot-test',
                'selfLink': 'https://www.googleapis.com/compute/v1/projects/spotty-221422/global/snapshots/snapshot-test',
                'sourceDisk': 'https://www.googleapis.com/compute/v1/projects/spotty-221422/zones/us-east1-b/disks/disk-test',
                'sourceDiskId': '599723469887162882',
                'status': 'READY',
                'storageBytes': '0',
                'storageBytesStatus': 'UP_TO_DATE',
                'storageLocations': ['us-central1']}
        """
        self._data = data

    @staticmethod
    def get_by_name(ce: CEClient, snapshot_name: str):
        """Returns a snapshot by its name."""
        res = ce.list_snapshots(snapshot_name)
        if not res:
            return None

        return Snapshot(res[0])

    @property
    def name(self) -> str:
        return self._data['name']

    @property
    def size(self) -> int:
        return self._data['diskSizeGb']

    @property
    def self_link(self) -> str:
        return self._data['selfLink']



import logging
import subprocess
from spotty.commands.writers.abstract_output_writrer import AbstractOutputWriter
from spotty.deployment.abstract_ssh_instance_manager import AbstractSshInstanceManager
from spotty.providers.remote.config.instance_config import InstanceConfig
from spotty.providers.remote.helpers.rsync import get_upload_command, check_rsync_installed, get_download_command


class InstanceManager(AbstractSshInstanceManager):

    instance_config: InstanceConfig

    def _get_instance_config(self, instance_config: dict) -> InstanceConfig:
        """Validates the instance config and returns an InstanceConfig object."""
        return InstanceConfig(instance_config, self.project_config)

    def is_running(self):
        """Assuming the remote instance is running."""
        return True

    def clean(self, output: AbstractOutputWriter):
        pass

    def sync(self, output: AbstractOutputWriter, dry_run=False):

        output.write('Syncing files with the instance...')

        # check rsync is installed
        check_rsync_installed()

        # sync the project with the instance
        rsync_cmd = get_upload_command(
            local_dir=self.project_config.project_dir,
            remote_dir=self.instance_config.host_project_dir,
            ssh_user=self.ssh_user,
            ssh_host=self.ssh_host,
            ssh_key_path=self.ssh_key_path,
            ssh_port=self.ssh_port,
            filters=self.project_config.sync_filters,
            use_sudo=(not self.instance_config.container_config.run_as_host_user),
            dry_run=dry_run,
        )

        # execute the command locally
        logging.debug('rsync command: ' + rsync_cmd)
        exit_code = subprocess.call(rsync_cmd, shell=True)
        if exit_code != 0:
            raise ValueError('Failed to upload files to the instance.')

    def download(self, download_filters: list, output: AbstractOutputWriter, dry_run=False):

        output.write('Downloading files from the instance...')

        # check rsync is installed
        check_rsync_installed()

        # sync the project with the instance
        rsync_cmd = get_download_command(
            local_dir=self.project_config.project_dir,
            remote_dir=self.instance_config.host_project_dir,
            ssh_user=self.ssh_user,
            ssh_host=self.ssh_host,
            ssh_key_path=self.ssh_key_path,
            ssh_port=self.ssh_port,
            filters=download_filters,
            use_sudo=(not self.instance_config.container_config.run_as_host_user),
            dry_run=dry_run,
        )

        # execute the command locally
        logging.debug('rsync command: ' + rsync_cmd)
        exit_code = subprocess.call(rsync_cmd, shell=True)
        if exit_code != 0:
            raise ValueError('Failed to download files from the instance.')

    @property
    def ssh_host(self) -> str:
        return self.instance_config.host

    @property
    def ssh_key_path(self) -> str:
        return self.instance_config.key_path

    @property
    def ssh_port(self) -> int:
        return self.instance_config.port
